import os
import numpy as np
from prometheus import utils_numba as utils, models_numba as mnb, getpsf_numba as gnb
from prometheus import groupfit_numba as gfit, allfit_numba as afit
from prometheus import ccddata

def tests():
    utils_tests()
    models_tests()
    getpsf_tests()

def utils_tests():

    ###############
    # UTILS
    ###############

    sm = utils.nansum(np.random.rand(100))
    sm = utils.nansum(np.random.rand(10,10))
    arr = np.random.rand(100)
    arr[10:20] = np.nan
    sm = utils.nansum(arr)
    print('utils_numba.nansum() okay')

    sm = utils.sum(np.random.rand(100))
    sm = utils.sum(np.random.rand(10,10))
    arr = np.random.rand(10,10)
    arr[0,1] = np.nan
    sm = utils.sum(arr,ignore_nan=True)
    print('utils_numba.sum() okay')

    sm = utils.sum2d(np.random.rand(10,10),axis=0)
    sm = utils.sum2d(np.random.rand(10,10),axis=1)
    arr = np.random.rand(10,10)
    arr[0,1] = np.nan
    sm = utils.sum2d(arr,axis=0,ignore_nan=True)
    print('utils_numba.sum2d() okay')

    sm = utils.sum3d(np.random.rand(10,10,20),axis=0)
    sm = utils.sum3d(np.random.rand(10,10,20),axis=1)
    sm = utils.sum3d(np.random.rand(10,10,20),axis=2)
    arr = np.random.rand(10,10,20)
    arr[0,1,2] = np.nan
    sm = utils.sum3d(arr,axis=0,ignore_nan=True)
    print('utils_numba.sum3d() okay')


    mn = utils.nanmean(np.random.rand(100))
    mn = utils.nanmean(np.random.rand(10,10))
    arr = np.random.rand(100)
    arr[10:20] = np.nan
    mn = utils.nanmean(arr)
    print('utils_numba.nanmean() okay')

    mn = utils.mean(np.random.rand(100))
    mn = utils.mean(np.random.rand(10,10))
    arr = np.random.rand(10,10)
    arr[0,1] = np.nan
    mn = utils.mean(arr,ignore_nan=True)
    print('utils_numba.mean() okay')

    mn = utils.mean2d(np.random.rand(10,10),axis=0)
    mn = utils.mean2d(np.random.rand(10,10),axis=1)
    arr = np.random.rand(10,10)
    arr[0,1] = np.nan
    mn = utils.mean2d(arr,axis=0,ignore_nan=True)
    print('utils_numba.mean2d() okay')

    mn = utils.mean3d(np.random.rand(10,10,20),axis=0)
    mn = utils.mean3d(np.random.rand(10,10,20),axis=1)
    mn = utils.mean3d(np.random.rand(10,10,20),axis=2)
    arr = np.random.rand(10,10,20)
    arr[0,1,2] = np.nan
    mn = utils.mean3d(arr,axis=0,ignore_nan=True)
    print('utils_numba.mean3d() okay')


    med = utils.nanmedian(np.random.rand(100))
    med = utils.nanmedian(np.random.rand(10,10))
    arr = np.random.rand(100)
    arr[10:20] = np.nan
    med = utils.nanmedian(arr)
    print('utils_numba.nanmedian() okay')

    med = utils.median(np.random.rand(100))
    med = utils.median(np.random.rand(10,10))
    arr = np.random.rand(10,10)
    arr[0,1] = np.nan
    med = utils.median(arr,ignore_nan=True)
    print('utils_numba.median() okay')

    med = utils.median2d(np.random.rand(10,10),axis=0)
    med = utils.median2d(np.random.rand(10,10),axis=1)
    arr = np.random.rand(10,10)
    arr[0,1] = np.nan
    med = utils.median2d(arr,axis=0,ignore_nan=True)
    print('utils_numba.median2d() okay')

    med = utils.median3d(np.random.rand(10,10,20),axis=0)
    med = utils.median3d(np.random.rand(10,10,20),axis=1)
    med = utils.median3d(np.random.rand(10,10,20),axis=2)
    arr = np.random.rand(10,10,20)
    arr[0,1,2] = np.nan
    med = utils.median3d(arr,axis=0,ignore_nan=True)
    print('utils_numba.median3d() okay')
    
    
    sig = utils.mad(np.random.rand(100))
    sig = utils.mad(np.random.rand(10,10))
    arr = np.random.rand(10,10,20)
    arr[0,1,2] = np.nan
    sig = utils.mad(arr,ignore_nan=True)
    print('utils_numba.mad() okay')

    sig = utils.mad2d(np.random.rand(10,10),axis=0)
    sig = utils.mad2d(np.random.rand(10,10),axis=1)
    arr = np.random.rand(10,10)
    arr[0,1] = np.nan
    sig = utils.mad2d(arr,axis=0,ignore_nan=True)
    print('utils_numba.mad2d() okay')

    sig = utils.mad3d(np.random.rand(10,10,20),axis=0)
    sig = utils.mad3d(np.random.rand(10,10,20),axis=1)
    sig = utils.mad3d(np.random.rand(10,10,20),axis=2)
    arr = np.random.rand(10,10,20)
    arr[0,1,2] = np.nan
    sig = utils.mad3d(arr,axis=0,ignore_nan=True)
    print('utils_numba.mad3d() okay')

    x = np.arange(1,4,1.0)
    y = 1.0+0.5*x-0.33*x**2
    out = utils.quadratic_bisector(x,y)
    print('utils_numba.quadratic_bisector() okay')

    x,y = utils.meshgrid(np.arange(10),np.arange(20))
    print('utils_numba.meshgrid() okay')
    
    out = utils.aclip(np.random.rand(100),-0.5,0.5)
    print('utils_numba.aclip() okay')
    
    out = utils.clip(-0.2,-0.5,0.5)
    print('utils_numba.clip() okay')
    
    out = utils.gamma(0.5)
    print('utils_numba.gamma() okay')
    
    out = utils.gammaincinv05(1.5)
    print('utils_numba.gammaincinv05() okay')

    data = np.random.rand(10,20)
    out = utils.linearinterp(data,1.4,2.5)
    print('utils_numba.linearinterp() okay')

    data = np.random.rand(10,20)
    x = np.array([1.4,2.5])
    y = np.array([2.6,3.4])
    out = utils.alinearinterp(data,x,y)
    print('utils_numba.alinearinterp() okay')

    arr = np.random.rand(20,20)
    out = utils.inverse(arr)
    print('utils_numba.inverse() okay')

    jac = np.random.rand(200,5)
    resid = np.random.rand(10,20).ravel()
    weight = np.ones((10,20),float).ravel()
    out = utils.qr_jac_solve(jac,resid,weight=weight)
    print('utils_numba.qr_jac_solve() okay')

    jac = np.random.rand(200,5)
    resid = np.random.rand(10,20).ravel()
    weight = np.ones((10,20),float).ravel()
    out = utils.jac_covariance(jac,resid,weight)
    print('utils_numba.jac_covariance() okay')

    pars = np.array([1.0,2.0,3.0])
    bounds = np.zeros((3,2),float)
    bounds[:,0] = [0.0,0.0,0.0]
    bounds[:,1] = [4.0,2.0,3.5]
    out = utils.checkbounds(pars,bounds)
    print('utils_numba.checkbounds() okay')

    pars = np.array([1.0,2.0,3.0])
    bounds = np.zeros((3,2),float)
    bounds[:,0] = [0.0,0.0,0.0]
    bounds[:,1] = [4.0,2.0,3.5]
    out = utils.limbounds(pars,bounds)
    print('utils_numba.limbounds() okay')

    steps = np.array([0.5,0.1,0.2])
    maxsteps = np.array([0.6,0.2,0.1])
    out = utils.limsteps(steps,maxsteps)
    print('utils_numba.limsteps() okay')

    pars = np.array([1.0,2.0,3.0])
    bounds = np.zeros((3,2),float)
    bounds[:,0] = [0.0,0.0,0.0]
    bounds[:,1] = [4.0,2.0,3.5]
    steps = np.array([0.5,0.1,0.2])
    maxsteps = np.array([0.6,0.2,0.1])
    out = utils.newpars(pars,steps,bounds=bounds,maxsteps=maxsteps)
    print('utils_numba.newpars() okay')

    xdata = np.zeros((10,2),float)
    xdata[:,0] = np.arange(10)
    xdata[:,1] = np.arange(10)
    pars = np.array([1.0,2.0,3.0,4.0])
    out = utils.poly2d(xdata,pars)
    print('utils_numba.poly2d() okay')
    
    out = utils.jacpoly2d(xdata,pars)
    print('utils_numba.jacpoly2d() okay')

    x,y = np.meshgrid(np.arange(10),np.arange(10))
    xdata = np.zeros((100,2),float)
    xdata[:,0] = x.ravel()
    xdata[:,1] = y.ravel()
    pars = np.array([1.0,2.0,3.0,4.0])
    data = utils.poly2d(xdata,pars)
    error = data*0+1
    out = utils.poly2dfit(xdata[:,0],xdata[:,1],data,error,maxiter=2,minpercdiff=0.5,verbose=False)
    print('utils_numba.poly2dfit() okay')

    # Index
    arr = np.array([1.0,2.0,3.0,4.0,5.0,3.0,2.0,1.0,6.0])
    index = utils.Index(arr)
    out = index.data
    out = index.index
    out = index.num
    out = len(index)
    out = index[0]
    out = index.values
    out = index.get(0)
    out = index.getindex(0)
    out = index.invindex
    print('utils_numba.Index okay')

    ra1 = np.random.rand(100)*10
    dec1 = np.random.rand(100)*10
    ra2 = ra1 + np.random.randn(100)*0.1
    dec2 = ra1 + np.random.randn(100)*0.1
    n1 = len(ra1)
    n2 = len(ra2)
    X1 = np.zeros((n1,2),float)
    X1[:,0] = ra1
    X1[:,1] = dec1
    X2 = np.zeros((n2,2),float)
    X2[:,0] = ra2
    X2[:,1] = dec2
    out = utils.crossmatch(X1, X2, max_distance=np.inf,k=1)
    print('utils_numba.crossmatch() okay')

    ra1 = np.random.rand(100)*10
    dec1 = np.random.rand(100)*10
    ra2 = ra1 + np.random.randn(100)*0.001
    dec2 = dec1 + np.random.randn(100)*0.001
    out = utils.xmatch(ra1, dec1, ra2, dec2, dcr=2.0, unique=False, sphere=False)
    out = utils.xmatch(ra1, dec1, ra2, dec2, dcr=2.0, unique=True, sphere=False)
    out = utils.xmatch(ra1, dec1, ra2, dec2, dcr=2.0, unique=False, sphere=True)
    out = utils.xmatch(ra1, dec1, ra2, dec2, dcr=2.0, unique=True, sphere=True)
    print('utils_numba.xmatch() okay')

    imshape = (1000,1000)
    multi_index = (np.random.choice(np.arange(100),100),
                   np.random.choice(np.arange(100),100),
                   np.random.choice(np.arange(100),100))
    out = utils.ravel_multi_index(multi_index,imshape)
    print('utils_numba.ravel_multi_index() okay')

    imshape = (1000,1000)
    indices = np.arange(5000)
    out = utils.unravel_index(indices,imshape)
    print('utils_numba.unravel_ndex() okay')

    arr = np.array([1.0,2.0,3.0,4.0,5.0,3.0,2.0,1.0,6.0])
    out = utils.unique_index(arr)
    print('utils_numba.unique_index() okay')


def models_tests():
    
    ###############
    # MODELS
    ###############
    
    # run individual model functions and check that they work without crashing

    # x,y = utils.meshgrid(np.arange(51),np.arange(51))
    # im = mnb.gaussian2d()
    # out = mnb.gaussfwhm(im)
    # print('models_numba.gaussfwhm() okay')
    
    # out = mnb.contourfwhm(im)
    # print('models_numba.contourfwhm() okay')
    
    # out = mnb.imfwhm(im)
    # print('models_numba.imfwhm() okay')
    
    #out = mnb.linearinterp(binim,fullim,binsize)
    #print('models_numba.linearinterp() okay')
    
    #out = mnb.checkbounds(pars,bounds)
    #print('models_numba.checkbounds() okay')
    
    #out = mnb.limbounds(pars,bounds)
    #print('models_numba.limbounds() okay')
    
    #out = mnb.limsteps(steps,maxsteps)
    #print('models_numba.limsteps() okay')
    
    # out = mnb.newlsqpars(pars,steps,bounds,maxsteps)
    # print('models_numba.newlsqpars() okay')
    
    # out = mnb.newbestpars(bestpars,dbeta)
    # print('models_numba.newbestpars() okay')
    
    # out = mnb.starbbox(coords,imshape,radius)
    # print('models_numba.starbbox() okay')
    
    # out = mnb.bbox2xy(bbox)
    # print('models_numba.bbo2xy() okay')
    
    out = mnb.gauss_abt2cxy(3.1,3.0,0.1)
    print('models_numba.gauss_abt2cxy() okay')
    
    out = mnb.gauss_cxy2abt(0.2,0.1,-0.1)
    print('models_numba.gauss_cxy2abt() okay')

    pars_gauss = np.array([100.0,3.5,4.5,3.1,3.0,0.1])
    out = mnb.gaussian2d_flux(pars_gauss)
    print('models_numba.gaussian2d_flux() okay')

    out = mnb.gaussian2d_fwhm(pars_gauss)
    print('models_numba.gaussian2d_fwhm() okay')

    out = mnb.gaussian2d(1.0,2.0,pars_gauss,6)
    print('models_numba.gaussian2d() okay')
    
    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    out = mnb.agaussian2d(x,y,pars_gauss,3)
    print('models_numba.agaussian2d() okay')

    # x,y = mnb.meshgrid(np.arange(51),np.arange(51))
    # pars = np.array([100.0,3.5,4.5,3.1,3.0,0.1])
    # im,deriv = mnb.agaussian2d(x,y,pars,3)
    # err = np.sqrt(np.maximum(im,1))
    # out = mnb.gaussian2dfit(im,err,90.0,3.1,4.1,False)
    # print('models_numba.gaussian2dfit() okay')

    pars_moffat = np.array([100.0,3.5,4.5,3.1,3.0,0.1,2.5])
    out = mnb.moffat2d_fwhm(pars_moffat)
    print('models_numba.moffat2d_fwhm() okay')

    pars = np.array([100.0,3.5,4.5,3.1,3.0,0.1,2.5])
    out = mnb.moffat2d_flux(pars)
    print('models_numba.moffat2d_flux() okay')

    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    out = mnb.amoffat2d(x,y,pars_moffat,7)
    print('models_numba.amoffat2d() okay')

    out = mnb.moffat2d(1.0,2.0,pars_moffat,7)
    print('models_numba.moffat2d() okay')

    # x,y = mnb.meshgrid(np.arange(51),np.arange(51))
    # pars = np.array([100.0,3.5,4.5,3.1,3.0,0.1,2.5])
    # im,deriv = mnb.amoffat2d(x,y,pars,7)
    # err = np.sqrt(np.maximum(im,1))
    # out = mnb.moffat2dfit(im,err,90.0,3.1,4.7,verbose)
    # print('models_numba.moffat2dfit() okay')

    pars_penny = np.array([100.0,5.55,6.33,3.1,3.0,0.1,0.1,6])
    out = mnb.penny2d_fwhm(pars_penny)
    print('models_numba.penny2d_fwhm() okay')

    out = mnb.penny2d_flux(pars_penny)
    print('models_numba.penny2d_flux() okay')

    out = mnb.apenny2d(x,y,pars_penny,8)
    print('models_numba.apenny2d() okay')

    out = mnb.penny2d(5.0,6.6,pars_penny,8)
    print('models_numba.penny2d() okay')

    # pars = np.array([100.0,5.55,6.33,3.1,3.0,0.1,0.1,6])
    # out = mnb.penny2dfit(im,err,ampc,xc,yc,verbose)
    # print('models_numba.penny2dfit() okay')

    pars_gausspow = np.array([100.0,5.55,6.33,5.0,7.0,0.01,0.9,0.8])
    out = mnb.gausspow2d_fwhm(pars_gausspow)
    print('models_numba.gausspow2d_fwhm() okay')

    out = mnb.gausspow2d_flux(pars_gausspow)
    print('models_numba.gausspow2d_flux() okay')

    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    out = mnb.agausspow2d(x,y,pars_gausspow,8)
    print('models_numba.agausspow2d() okay')

    out = mnb.gausspow2d(5.5,6.5,pars_gausspow,8)
    print('models_numba.gausspow2d() okay')

    # pars = np.array([100.0,5.55,6.33,5.0,7.0,0.01,0.9,0.8])
    # out = mnb.gausspow2dfit(im,err,ampc,xc,yc,verbose)
    # print('models_numba.gausspow2dfit() okay')

    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    pars_sersic = np.array([100.0,5.5,6.3,0.3,1.0,0.9,0.0])
    out = mnb.asersic2d(x,y,pars_sersic,7)
    print('models_numba.asersic2d() okay')

    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    out = mnb.sersic2d(5.5, 6.5, pars_sersic,7)
    print('models_numba.sersic2d() okay')

    out = mnb.sersic2d_fwhm(pars_sersic)
    print('models_numba.sersic2d_fwhm() okay')

    out = mnb.sersic_b(1.5)
    print('models_numba.sersic_b() okay')
    
    # out = mnb.create_sersic_function(100.0, re, 1.0)
    # print('models_numba.create_sersic_function() okay')
    
    # out = mnb.sersic_lum(100.0, re, 1.0)
    # print('models_numba.sersic_lum() okay')
    
    # out = mnb.sersic_full2half(100.0,kserc,alpha)
    # print('models_numba.sersic_full2half() okay')
    
    # out = mnb.sersic_half2full(100.0,Re,alpha)
    # print('models_numba.sersic_half2ful() okay')

    out = mnb.sersic2d_flux(pars_sersic)
    print('models_numba.sersic2d_flux() okay')

    # not njit yet
    out = mnb.sersic2d_estimates(pars_sersic)
    print('models_numba.sersic2d_estimates() okay')

    out = mnb.model2d(5.5,6.5,1,pars_gauss,6)
    out = mnb.model2d(5.5,6.5,2,pars_moffat,7)
    out = mnb.model2d(5.5,6.5,3,pars_penny,8)
    out = mnb.model2d(5.5,6.5,4,pars_gausspow,8)
    out = mnb.model2d(5.5,6.5,5,pars_sersic,7)
    print('models_numba.model2d() okay')

    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    out = mnb.amodel2d(x,y,1,pars_gauss,6)
    out = mnb.amodel2d(x,y,2,pars_moffat,7)
    out = mnb.amodel2d(x,y,3,pars_penny,8)
    out = mnb.amodel2d(x,y,4,pars_gausspow,8)
    out = mnb.amodel2d(x,y,5,pars_sersic,7)
    print('models_numba.amodel2d() okay')

    out = mnb.model2d_flux(1,pars_gauss)
    out = mnb.model2d_flux(2,pars_moffat)
    out = mnb.model2d_flux(3,pars_penny)
    out = mnb.model2d_flux(4,pars_gausspow)
    out = mnb.model2d_flux(5,pars_sersic)
    print('models_numba.model2d_flux() okay')

    out = mnb.model2d_fwhm(1,pars_gauss)
    out = mnb.model2d_fwhm(2,pars_moffat)
    out = mnb.model2d_fwhm(3,pars_penny)
    out = mnb.model2d_fwhm(4,pars_gausspow)
    out = mnb.model2d_fwhm(5,pars_sersic)
    print('models_numba.model2d_fwhm() okay')

    out = mnb.model2d_estimates(1,100.0,5.5,6.1)
    out = mnb.model2d_estimates(2,100.0,5.5,6.1)
    out = mnb.model2d_estimates(3,100.0,5.5,6.1)
    out = mnb.model2d_estimates(4,100.0,5.5,6.1)
    out = mnb.model2d_estimates(5,100.0,5.5,6.1)
    print('models_numba.model2d_estimates() okay')
    
    out = mnb.model2d_bounds(1)
    out = mnb.model2d_bounds(2)
    out = mnb.model2d_bounds(3)
    out = mnb.model2d_bounds(4)
    out = mnb.model2d_bounds(5)
    print('models_numba.model2d_bounds() okay')

    out = mnb.model2d_maxsteps(1,pars_gauss)
    out = mnb.model2d_maxsteps(2,pars_moffat)
    out = mnb.model2d_maxsteps(3,pars_penny)
    out = mnb.model2d_maxsteps(4,pars_gausspow)
    out = mnb.model2d_maxsteps(5,pars_sersic)
    print('models_numba.model2d_maxsteps() okay')

    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    im,_ = mnb.amodel2d(x,y,1,pars_gauss,6)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.model2dfit(im,err,x,y,1,90.0,5.1,6.2,False)
    
    im,_ = mnb.amodel2d(x,y,2,pars_moffat,7)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.model2dfit(im,err,x,y,2,90.0,5.1,6.2,False)
    
    im,_ = mnb.amodel2d(x,y,3,pars_penny,8)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.model2dfit(im,err,x,y,3,90.0,5.1,6.2,False)
    
    im,_ = mnb.amodel2d(x,y,4,pars_gausspow,8)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.model2dfit(im,err,x,y,4,90.0,5.1,6.2,False)

    im,_ = mnb.amodel2d(x,y,5,pars_sersic,7)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.model2dfit(im,err,x,y,5,90.0,5.1,6.2,False)
    print('models_numba.model2dfit() okay')
    
    # out = mnb.relcoord(x,y,shape)
    # print('models_numba.relcoord() okay')
    
    # out = mnb.empirical(x, y, pars, data, imshape=None, deriv=False)
    # print('models_numba.empirical() okay')

    psf = mnb.packpsf(1,pars_gauss[3:],lookup=np.zeros((11,11,4),float),imshape=(1000,1000))
    print('models_numba.packpsf() okay')
    
    # psftype,pars,npsfx,npsfy,psforder,nxhalf,nyhalf,lookup
    out = mnb.unpackpsf(psf)
    print('models_numba.unpackpsf() okay')
    
    out = mnb.psfinfo(psf)
    print('models_numba.psfinfo() okay')

    out = mnb.psf2d_fwhm(psf)
    print('models_numba.psf2d_fwhm() okay')

    out = mnb.psf2d_flux(psf,100.0,5.5,6.6)
    print('models_numba.psf2d_flux() okay')

    out = mnb.psf2d(x,y,psf,100.0,5.5,6.6,deriv=False,verbose=False)
    print('models_numba.psf2d() okay')

    im,_ = mnb.psf2d(x,y,psf,100.0,5.5,6.6)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.psf2dfit(im,err,x,y,psf,90.0,5.1,6.1,verbose=False)
    print('models_numba.psf2dfit() okay')
    
    pars = np.array([1.00,5.5,6.6])
    lookup = np.zeros((1,1,1),float)
    imshape = (1000,1000)
    out = mnb.psf(x,y,pars,1,pars_gauss[3:],lookup,imshape,deriv=False,verbose=False)
    out = mnb.psf(x,y,pars,2,pars_moffat[3:],lookup,imshape,deriv=False,verbose=False)
    out = mnb.psf(x,y,pars,3,pars_penny[3:],lookup,imshape,deriv=False,verbose=False)
    out = mnb.psf(x,y,pars,4,pars_gausspow[3:],lookup,imshape,deriv=False,verbose=False)
    out = mnb.psf(x,y,pars,5,pars_sersic[3:],lookup,imshape,deriv=False,verbose=False)
    print('models_numba.psf() okay')

    
    x,y = utils.meshgrid(np.arange(51),np.arange(51))
    pars = np.array([90.0,5.1,6.2])
    lookup = np.zeros((1,1,1),float)
    imshape = (1000,1000)
    im,_ = mnb.psf(x,y,pars,1,pars_gauss[3:],lookup,imshape,deriv=False,verbose=False)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.psffit(im,err,x,y,pars,1,pars_gauss[3:],lookup,imshape,False)

    im,_ = mnb.psf(x,y,pars,2,pars_moffat[3:],lookup,imshape,deriv=False,verbose=False)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.psffit(im,err,x,y,pars,2,pars_moffat[3:],lookup,imshape,False)

    im,_ = mnb.psf(x,y,pars,3,pars_penny[3:],lookup,imshape,deriv=False,verbose=False)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.psffit(im,err,x,y,pars,3,pars_penny[3:],lookup,imshape,False)

    im,_ = mnb.psf(x,y,pars,4,pars_gausspow[3:],lookup,imshape,deriv=False,verbose=False)
    err = np.sqrt(np.maximum(im,1))
    out = mnb.psffit(im,err,x,y,pars,4,pars_gausspow[3:],lookup,imshape,False)

    im,_ = mnb.psf(x,y,pars,5,pars_sersic[3:],lookup,imshape,deriv=False,verbose=False)
    err = np.sqrt(np.maximum(im,1))
    #out = mnb.psffit(im,err,x,y,pars,5,pars_sersic[3:],lookup,imshape,False)
    # ZeroDivisionError: division by zero
    print('models_numba.psffit() okay')
    

    bbox = mnb.BoundingBox(10,20,30,40)
    print('models_numba.BoundingBox.__init__() okay')
    
    out = bbox.xrange
    print('models_numba.BoundingBox.xrange okay')

    out = bbox.yrange
    print('models_numba.BoundingBox.yrange okay')

    out = bbox.yrange
    print('models_numba.BoundingBox.yrange okay')

    out = bbox.data
    print('models_numba.BoundingBox.data okay')

    im = np.random.rand(100,100)
    out = bbox.slice(im)
    print('models_numba.BoundingBox.slice() okay')

    out = bbox[0]
    print('models_numba.BoundingBox.__getitem__() okay')

    out = bbox.xy()
    print('models_numba.BoundingBox.xy() okay')

    out = bbox.reset()
    print('models_numba.BoundingBox.reset() okay')

    
    psf = mnb.PSFGaussian(pars_gauss[3:])
    print('models_numba.PSFGaussian.__init__() okay')

    out = psf.params
    print('models_numba.PSFGaussian.params okay')

    psf.params[0] = 1.0
    print('models_numba.PSFGaussian.prarams setter okay')

    out = psf.haslookup
    print('models_numba.PSFGaussian.haslookup okay')

    out = psf.starbbox((5.5,6.5),(1000,1000), 5.5)
    print('models_numba.PSFGaussian.starbbox() okay')

    out = psf.unitfootflux
    print('models_numba.PSFGaussian.unitfootflux okay')

    out = psf.fwhm()
    print('models_numba.PSFGaussian.fwhm() okay')

    out = psf.flux()
    print('models_numba.PSFGaussian.flux() okay')

    out = psf.evaluate(x,y,pars_gauss)
    print('models_numba.PSFGaussian.evaluate() okay')

    out = psf.deriv(x,y,pars_gauss)
    print('models_numba.PSFGaussian.deriv() okay')

    
    psf = mnb.PSF(1,pars_gauss[3:])
    print('models_numba.PSF.__init__() okay')

    out = psf.nparams
    print('models_numba.PSF.nparams okay')

    out = psf.params
    print('models_numba.PSF.params okay')

    psf.params[0] = 1.0
    print('models_numba.PSF.params setter okay')
    
    out = psf.name
    print('models_numba.PSF.name okay')
    
    out = psf.haslookup
    print('models_numba.PSF.haslookup okay')

    out = psf.starbbox((5.5,6.6),(1000,1000),5.5)
    print('models_numba.PSF.starbbox() okay')

    out = str(psf)
    print('models_numba.PSF.__str__() okay')

    out = psf.fwhm()
    print('models_numba.PSF.fwhm() okay')

    out = psf.flux()
    print('models_numba.PSF.flux() okay')

    pars = np.array([100.0,5.5,6.5])
    out = psf.evaluate(x,y,pars)
    print('models_numba.PSF.evaluate() okay')

    out = psf.model(x,y,pars)
    print('models_numba.PSF.model() okay')
    
    out = psf.deriv(x,y,pars)
    print('models_numba.PSF.deriv() okay')

    #out = psf.packpsf()
    #print('models_numba.PSF.packpsf() okay')
    
    
def getpsf_tests():
    
    ###############
    # GETPSF
    ###############

    out = gnb.starcube(tab,image,error,npix=51,fillvalue=np.nan)
    print('getpsf_numba.starcube() okay')

    out = gnb.mkempirical(cube,order=0,coords=None,shape=None,lookup=False)
    print('getpsf_numba.mkempirical() okay')

    out = gnb.starbbox((5.5,6.5),(1000,1000),5.6)
    print('getpsf_numba.starbbox() okay')

    #out = gnb.sliceinsert(array,lo,insert)
    #print('getpsf_numba.sliceinsert() okay')

    out = gnb.getstar(image,error,xcen,ycen,fitradius)
    print('getpsf_numba.getstar() okay')

    out = gnb.collatestars(image,error,starx,stary,fitradius)
    print('getpsf_numba.collatestars() okay')

    out = gnb.unpackstar(imdata,errdata,xdata,ydata,bbox,shape,istar)
    print('getpsf_numba.unpackstar() okay')

    out = gnb.getfitstar(image,error,xcen,ycen,fitradius)
    print('getpsf_numba.getfitstar() okay')

    out = gnb.collatefitstars(image,error,starx,stary,fitradius)
    print('getpsf_numba.collatefitstars() okay')

    out = gnb.unpackfitstar(imdata,errdata,xdata,ydata,bbox,ndata,istar)
    print('getpsf_numba.unpackfitstar() okay')

    out = gnb.PSFFitter()
    print('getpsf_numba.PSFFitter.__init__() okay')

    out = gnb.PSFFitter.unpackstar()
    print('getpsf_numba.PSFFitter.unpackstar() okay')

    out = gnb.PSFFitter.unpackfitstar()
    print('getpsf_numba.PSFFitter.unpackfitstar() okay')

    out = gnb.PSFFitter.psf()
    print('getpsf_numba.PSFFitter.psf() okay')

    out = gnb.PSFFitter.model()
    print('getpsf_numba.PSFFitter.model() okay')

    out = gnb.PSFFitter.chisq()
    print('getpsf_numba.PSFFitter.chisq() okay')

    out = gnb.PSFFitter.fitstars()
    print('getpsf_numba.PSFFitter.fitstars() okay')

    out = gnb.PSFFitter.jac()
    print('getpsf_numba.PSFFitter.jac() okay')

    out = gnb.PSFFitter.linesearch()
    print('getpsf_numba.PSFFitter.linesearch() okay')
    
    out = gnb.PSFFitter.starmodel()
    print('getpsf_numba.PSFFitter.starmodel() okay')

    out = gnb.fitpsf(psftype,psfparams,image,error,starx,stary,starflux,fitradius,method='qr',maxiter=10,
                     minpercdiff=1.0,verbose=False)
    print('getpsf_numba.fitpsf() okay')

    out = gnb.getpsf(psf,image,tab,fitradius=None,lookup=False,lorder=0,method='qr',subnei=False,
                     alltab=None,maxiter=10,minpercdiff=1.0,reject=False,maxrejiter=3,verbose=False)
    print('getpsf_numba.getpsf() okay')


def groupfit_tests():
    """  Testing the groupfit code."""

    psftab = np.zeros((4,4),np.float64)
    psftab[:,0] = np.arange(4)+1
    psftab[:,1] = [100,200,300,400]   # amp
    psftab[:,2] = [10,11,19,18]       # xcen
    psftab[:,3] = [20,30,31,21]       # ycen
    mpars = np.array([3.1,3.0,0.1])
    psf = mnb.PSF(1,mpars,21)
    xx,yy = np.meshgrid(np.arange(51),np.arange(51))
    model = np.zeros(51*51,float)
    for i in range(4):
        model += psf.model(xx,yy,psftab[i,1:])
    model = model.reshape(51,51)

    err = np.sqrt(model+10)
    sky = 0.0 #10.0
    im = model + sky + np.random.rand(51,51)*err
    mask = np.zeros(im.shape,bool)
    image = ccddata.CCDData(im,error=err,mask=mask)
    
    # initial estimates
    objtab = np.zeros((4,4),np.float64)
    objtab[:,0] = np.arange(4)+1
    objtab[:,1] = [80,250,200,500]           # amp
    objtab[:,2] = [9.5,11.6,19.2,17.8]       # xcen
    objtab[:,3] = [20.3,29.7,30.9,21.2]      # ycen

    gf = gfit.GroupFitter(psf.psftype,psf.params,im,err,objtab,3.0,
                          np.zeros((1,1,1),np.float64),psf.npix,True)
    print('groupfit.GroupFitter.__init__() okay')

    #out = gf.ampfit()
    #print('groupfit.GroupFitter.ampfit() okay')

    #out = gf.centroid()
    #print('groupfit.GroupFitter.centroid() okay')

    pars = np.zeros(7,np.float64)
    pars[:] = [100.0,3.4,5.5, 200.0,4.6,7.5, 100.0]
    bounds = [np.zeros(7,np.float64)-np.inf,np.zeros(7,np.float64)+np.inf]
    out = gf.checkbounds(pars,bounds)
    print('groupfit.GroupFitter.checkbounds() okay')

    out = gf.chisq(gf.pars)
    print('groupfit.GroupFitter.chisq() okay')

    out = gf.cov()
    print('groupfit.GroupFitter.cov() okay')

    out = gf.getstar(0)
    print('groupfit.GroupFitter.getstar() okay')

    out = gf.getstarfit(0)
    print('groupfit.GroupFitter.getstarfit() okay')

    out = gf.jac(gf.pars)
    print('groupfit.GroupFitter.jac() okay')

    pars = np.zeros(7,np.float64)
    pars[:] = [100.0,3.4,5.5, 200.0,4.6,7.5, 100.0]
    bounds = [np.zeros(7,np.float64)-np.inf,np.zeros(7,np.float64)+np.inf]
    out = gf.limbounds(pars,bounds)
    print('groupfit.GroupFitter.limbounds() okay')

    steps = np.zeros(6,np.float64)+0.1
    maxsteps = np.zeros(6,np.float64)+0.5
    out = gf.limsteps(steps,maxsteps)
    print('groupfit.GroupFitter.limsteps() okay')

    m,j = gf.jac(gf.pars)
    dbeta = utils.qr_jac_solve(j,gf.resflat)
    out = gf.linesearch(gf.pars,dbeta,m,j)
    print('groupfit.GroupFitter.linesearch() okay')

    out = gf.mkbounds(gf.pars,gf.imshape)
    print('groupfit.GroupFitter.mkbounds() okay')

    out = gf.model(gf.pars)
    print('groupfit.GroupFitter.model() okay')

    out = gf.modelstar(0)
    print('groupfit.GroupFitter.modelstar() okay')

    out = gf.modelstarfit(0)
    print('groupfit.GroupFitter.modelstarfit() okay')

    bounds = [np.zeros(13,np.float64)-np.inf,np.zeros(13,np.float64)+np.inf]
    steps = np.zeros(13,np.float64)+0.1
    maxsteps = np.zeros(13,np.float64)+0.5
    out = gf.newpars(gf.pars,steps,bounds,maxsteps)
    print('groupfit.GroupFitter.newpars() okay')

    out = gf.psf(gf.starx(0),gf.stary(0),gf.pars[:3])
    print('groupfit.GroupFitter.psf() okay')

    out = gf.psfjac(gf.starx(0),gf.stary(0),gf.pars[:3])
    print('groupfit.GroupFitter.psfjac() okay')

    #out = gf.score(0)
    #print('groupfit.GroupFitter.score() okay')

    out = gf.sky()
    print('groupfit.GroupFitter.sky() okay')

    out = gf.starbbox(0)
    print('groupfit.GroupFitter.starbbox() okay')

    out = gf.starfitbbox(0)
    print('groupfit.GroupFitter.starfitbbox() okay')

    out = gf.starfitchisq(0)
    print('groupfit.GroupFitter.starfitchisq() okay')

    out = gf.starfitinvindex(0)
    print('groupfit.GroupFitter.starfitinvindex() okay')

    out = gf.starfitnpix(0)
    print('groupfit.GroupFitter.starfitnpix() okay')

    out = gf.starfitravelindex(0)
    print('groupfit.GroupFitter.starfitravelindex() okay')

    out = gf.starfitrms(0)
    print('groupfit.GroupFitter.starfitrms() okay')

    out = gf.starfitx(0)
    print('groupfit.GroupFitter.starfitx() okay')

    out = gf.starfity(0)
    print('groupfit.GroupFitter.starfity() okay')

    out = gf.starflatindex(0)
    print('groupfit.GroupFitter.starflatindex() okay')

    out = gf.starflatnpix(0)
    print('groupfit.GroupFitter.starflatnpix() okay')

    out = gf.starflatx(0)
    print('groupfit.GroupFitter.starflatx() okay')

    out = gf.starflaty(0)
    print('groupfit.GroupFitter.starflaty() okay')

    out = gf.starnpix(0)
    print('groupfit.GroupFitter.starnpix() okay')
    
    out = gf.starravelindex(0)
    print('groupfit.GroupFitter.starravelindex() okay')

    out = gf.starx(0)
    print('groupfit.GroupFitter.starx() okay')

    out = gf.stary(0)
    print('groupfit.GroupFitter.stary() okay')

    out = gf.steps(gf.pars)
    print('groupfit.GroupFitter.steps() okay')

    out = gf.unfreeze()
    print('groupfit.GroupFitter.unfreeze() okay')
    
    out = gfit.fit(psf,im,err,objtab,verbose=False)
    print('groupfit.fit() okay')

    
def allfit_tests():
    """  Testing the allfit code."""

    psftab = np.zeros((4,4),np.float64)
    psftab[:,0] = np.arange(4)+1
    psftab[:,1] = [100,200,300,400]   # amp
    psftab[:,2] = [10,11,19,18]       # xcen
    psftab[:,3] = [20,30,31,21]       # ycen
    mpars = np.array([3.1,3.0,0.1])
    psf = mnb.PSF(1,mpars,21)
    xx,yy = np.meshgrid(np.arange(51),np.arange(51))
    model = np.zeros(51*51,float)
    for i in range(4):
        model += psf.model(xx,yy,psftab[i,1:])
    model = model.reshape(51,51)

    err = np.sqrt(model+10)
    sky = 0.0 #10.0
    im = model + sky + np.random.rand(51,51)*err
    mask = np.zeros(im.shape,bool)
    image = ccddata.CCDData(im,error=err,mask=mask)
    
    # initial estimates
    objtab = np.zeros((4,4),np.float64)
    objtab[:,0] = np.arange(4)+1
    objtab[:,1] = [80,250,200,500]           # amp
    objtab[:,2] = [9.5,11.6,19.2,17.8]       # xcen
    objtab[:,3] = [20.3,29.7,30.9,21.2]      # ycen

    fitradius = 3.0
    psflookup = np.zeros((1,1,1),np.float64)
    verbose = False
    nofreeze = False
    af = afit.AllFitter(psf.psftype,psf.params,psf.npix,psflookup,
                        image.data,image.error,image.mask,objtab,
                        fitradius,verbose,nofreeze)
    print('allfit.AllFitter.__init__() okay')

    pars = np.zeros(7,np.float64)
    pars[:] = [100.0,3.4,5.5, 200.0,4.6,7.5, 100.0]
    bounds = [np.zeros(7,np.float64)-np.inf,np.zeros(7,np.float64)+np.inf]
    out = af.checkbounds(pars,bounds)
    print('allfit.AllFitter.checkbounds() okay')

    out = af.chisq()
    print('allfit.AllFitter.chisq() okay')

    out = af.collatestars(af.imshape,af.starxcen,af.starycen,af.npix//2,af.fitradius,af.skyradius)
    print('allfit.AllFitter.collatestars() okay')

    out = af.fit()
    print('allfit.AllFitter.fit() okay')

    out = af.getstar(af.imshape,af.starxcen[0],af.starycen[0],af.npix//2,af.fitradius,af.skyradius)
    print('allfit.AllFitter.getstar() okay')

    pars = np.zeros(6,np.float64)
    pars[:] = [100.0,3.4,5.5, 200.0,4.6,7.5]
    bounds = [np.zeros(6,np.float64)-np.inf,np.zeros(6,np.float64)+np.inf]
    out = af.limbounds(pars,bounds)
    print('allfit.AllFitter.limbounds() okay')

    steps = np.zeros(6,np.float64)+0.1
    maxsteps = np.zeros(6,np.float64)+0.5
    out = af.limsteps(steps,maxsteps)
    print('allfit.AllFitter.limsteps() okay')

    out = af.mkbounds(af.pars,af.imshape)
    print('allfit.AllFitter.mkbounds() okay')

    bounds = [np.zeros(12,np.float64)-np.inf,np.zeros(12,np.float64)+np.inf]
    steps = np.zeros(12,np.float64)+0.1
    maxsteps = np.zeros(12,np.float64)+0.5
    out = af.newpars(af.pars,steps,bounds,maxsteps)
    print('allfit.AllFitter.newpars() okay')

    pars1,xind1,yind1,ravelindex1 = af.stardata(0)
    out = af.psf(xind1,yind1,pars1)
    print('allfit.AllFitter.psf() okay')

    pars1,xind1,yind1,ravelindex1 = af.stardata(0)
    out = af.psfjac(xind1,yind1,pars1)
    print('allfit.AllFitter.psfjac() okay')

    out = af.sky()
    print('allfit.AllFitter.sky() okay')
    
    out = af.starcov(0)
    print('allfit.AllFitter.starcov() okay')

    out = af.stardata(0)
    print('allfit.AllFitter.stardata() okay')

    out = af.starfit(0)
    print('allfit.AllFitter.starfit() okay')
    
    out = af.starfitchisq(0)
    print('allfit.AllFitter.starfitchisq() okay')

    out = af.starfitdata(0)
    print('allfit.AllFitter.starfitdata() okay')

    out = af.starfiterr(0)
    print('allfit.AllFitter.starfiterr() okay')

    out = af.starfitim(0)
    print('allfit.AllFitter.starfitim() okay')

    out = af.starfitnpix(0)
    print('allfit.AllFitter.starfitnpix() okay')

    out = af.starfitravelindex(0)
    print('allfit.AllFitter.starfitravelindex() okay')

    out = af.starfitresid(0)
    print('allfit.AllFitter.starfitresid() okay')

    out = af.starfitrms(0)
    print('allfit.AllFitter.starfitrms() okay')

    out = af.starjac(0)
    print('allfit.AllFitter.starjac() okay')

    out = af.starmodel(0)
    print('allfit.AllFitter.starmodel() okay')

    out = af.starmodelfull(0)
    print('allfit.AllFitter.starmodelfull() okay')

    out = af.starnpix(0)
    print('allfit.AllFitter.starnpix() okay')

    out = af.starravelindex(0)
    print('allfit.AllFitter.starravelindex() okay')

    out = af.starsky(0)
    print('allfit.AllFitter.starsky() okay')

    out = af.steps(af.pars)
    print('allfit.AllFitter.steps() okay')

    out = af.unfreeze()
    print('allfit.AllFitter.unfreeze() okay')


    out = afit.numba_allfit(psf.psftype,psf.params,psf.npix,psflookup,psf.flux(),
                            image.data,image.error,image.mask,objtab,fitradius,maxiter=10,
                            minpercdiff=0.5,reskyiter=2,verbose=False,
                            nofreeze=False)
    print('allfit.numba_allfit() okay')

    out = afit.allfit(psf,image,objtab)
    print('allfit.allfit() okay')

    #out = afit.fit(psf,image,objtab)
    #print('allfit.fit() okay')



""" CPU-time returning clock() function which works from within njit-ted code """
# from https://github.com/numba/numba/issues/4003
import ctypes
#import platform
#
#if platform.system() == "Windows":
#    from ctypes.util import find_msvcrt
#
#    __LIB = find_msvcrt()
#    if __LIB is None:
#        __LIB = "msvcrt.dll"
#else:
#    from ctypes.util import find_library
#
#    __LIB = find_library("c")
#
#clock = ctypes.CDLL(__LIB).clock
#clock.argtypes = []
#clock.restype = ctypes.c_int64

# https://github.com/open-atmos/PyMPDATA/blob/main/PyMPDATA/impl/clock.py
clock = ctypes.pythonapi._PyTime_GetSystemClock  # pylint:disable=protected-access
clock.argtypes = []
clock.restype = ctypes.c_int64

# returns time in nanoseconds
# divide time difference by 1e9 to get seconds



cpdef double[:,:] transpose(double[:,:] A)
cpdef double[:,:] matmult(double[:,:] A, double[:,:] B)
cpdef double[:] sub1d(double[:] arr1,double[:] arr2)
cpdef double[:] add1d(double[:] arr1,double[:] arr2)
cpdef double[:] mult1d(double[:] arr1,double[:] arr2)
cpdef double[:] pow1d(double[:] arr,double power)
cpdef double[:] fact1d(double[:] arr,double factor)
cpdef double[:,:] sub2d(double[:,:] arr1,double[:,:] arr2)
cpdef double[:,:] add2d(double[:,:] arr1,double[:,:] arr2)
cpdef double[:,:] mult2d(double[:,:] arr1,double[:,:] arr2)
cpdef double[:,:] pow2d(double[:,:] arr,double power)
cpdef double[:,:] fact2d(double[:,:] arr,double factor)
cpdef double[:,:,:] sub3d(double[:,:,:] arr1,double[:,:,:] arr2)
cpdef double[:,:,:] add3d(double[:,:,:] arr1,double[:,:,:] arr2)
cpdef double[:,:,:] mult3d(double[:,:,:] arr1,double[:,:,:] arr2)
cpdef double[:,:,:] pow3d(double[:,:,:] arr,double power)
cpdef double[:,:,:] fact3d(double[:,:,:] arr,double factor)
cpdef double[:] ravel2d(double[:,:] datain)
cpdef double[:] ravel3d(double[:,:,:] datain)
cpdef double[:,:] reshape2d(double[:,:] datain, Py_ssize_t[:] shapeout)
cpdef double[:,:,:] reshape3d(double[:,:,:] datain, Py_ssize_t[:] shapeout)
cdef double linearinterp(double[:,:] data, double x, double y)
cdef double[:] alinearinterp(double[:,:] data, double[:] x, double[:] y)
cpdef double nanmean(double[:] data)
cpdef tuple nonnanindexes(double[:] data)
cpdef double[:] nonnanarray(double[:] data)
cpdef double nanmedian(double[:] data)
cpdef double mad(double[:] data, int ignore_nan, int zero)
cpdef double[:] mad2d(double[:,:] data, int axis, int ignore_nan, int zero)
cpdef double[:,:] mad3d(double[:,:,:] data, int axis, int ignore_nan, int zero)
cpdef double quadratic_bisector(double[:] x, double[:] y)
cpdef long[:] arange(int start, int stop, int step)
cpdef double[:] linspace(double start, double stop, int num)
cdef list intmeshgrid(long[:] x, long[:] y)
cdef list meshgrid(double[:] x, double[:] y)
cdef double[:] aclip(double[:] val,double minval,double maxval)
cdef double clip(double val,double minval,double maxval)
cpdef Py_ssize_t[:] starbbox(double[:] coords,Py_ssize_t[:] imshape,double radius)
cpdef double[:,:] slice2d(double[:,:] array,Py_ssize_t[:] slc)
cpdef double[:,:] inverse(double[:,:] a)
cpdef double[:] qr_jac_solve(double[:,:] jac,double[:] resid,double[:] weight)
cpdef double[:,:] jac_covariance(double[:,:] jac,double[:] resid, double[:] weight)
cpdef double[:] poly2d(double[:,:] xdata, double[:] pars)
cpdef list jacpoly2d(double[:,:] xdata, double[:] pars)
cpdef list poly2dfit(double[:] x,double[:] y,double[:] data,double[:] error,int maxiter,double minpercdiff)
cpdef list relcoord(double[:] x, double[:] y, Py_ssize_t[:] shape)


#!/usr/bin/env python

"""GROUPFIT.PY - Fit groups of stars in an image

"""

__authors__ = 'David Nidever <dnidever@montana.edu?'
__version__ = '20210826'  # yyyymmdd


import numpy as np
from numba import njit,types,from_dtype
from numba.experimental import jitclass
from numba_kdtree import KDTree
#from . import models_numba as mnb, utils_numba as utils, getpsf_numba as gnb

from clock_numba import clock
import models_numba_static as mnb, utils_numba_static as utils
#getpsf_numba as gnb

# Fit a PSF model to multiple stars in an image

from numba.pycc import CC
cc = CC('_groupfit_numba_static')


# xdata = xind,yind
xdatatype = 'UniTuple(i8[:],2)'
# psfdata = (psftype,psfparams,psflookup,psforder,imshape)
psfdatatype = 'Tuple((i8,f8[:],f8[:,:,:],i8,UniTuple(i8,2)))'
# freezedata = (freezepars,freezestars)
freezedatatype = 'Tuple((b1[:],b1[:]))'
# flatdata = (starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix)
flatdatatype = 'Tuple((i8[:],i8[:,:],i8[:],i8[:],i8[:],i8))'
# stardata = (starravelindex,starndata,xx,yy)
stardatatype = 'Tuple((i8[:,:],i4[:],i8[:,:],i8[:,:]))'
# covflatdata = (starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix,imflat,errflat,skyflat)
covflatdatatype = 'Tuple((i8[:],i8[:,:],i8[:],i8[:],i8[:],i8,f8[:],f8[:],f8[:]))'

@njit
@cc.export('getstarinfo', '(UniTuple(i8,2),b1[:,:],f8,f8,i4,f8,f8)')
def getstarinfo(imshape,mask,xcen,ycen,hpsfnpix,fitradius,skyradius):
    """ Return a star's full footprint, fitted pixels, and sky pixels data."""
    # always return the same size
    # a distance of 6.2 pixels spans 6 full pixels but you could have
    # 0.1 left on one side and 0.1 left on the other side
    # that's why we have to add 2 pixels
    maxpix = int(3.14*skyradius**2)
    nskypix = int(2*skyradius+1)
    skybbox = utils.starbbox((xcen,ycen),imshape,int(skyradius))
    nsx = skybbox[1]-skybbox[0]
    nsy = skybbox[3]-skybbox[2]
    skyravelindex = np.zeros(maxpix,np.int64)-1
    nfpix = hpsfnpix*2+1
    #fbbox = utils.starbbox((xcen,ycen),imshape,hpsfnpix)
    # extra buffer is ALWAYS at the end of each dimension    
    fravelindex = np.zeros(nfpix*nfpix,np.int64)-1
    # Fitting pixels
    npix = int(np.floor(2*fitradius))+2
    bbox = utils.starbbox((xcen,ycen),imshape,fitradius)
    nx = bbox[1]-bbox[0]
    ny = bbox[3]-bbox[2]
    ravelindex = np.zeros(npix*npix,np.int64)-1
    fcount = 0
    count = 0
    skycount = 0
    for j in range(nskypix):
        y = j + skybbox[2]
        for i in range(nskypix):
            x = i + skybbox[0]
            if mask[y,x]:   # exclude bad pixels
                continue
            r = np.sqrt((x-xcen)**2 + (y-ycen)**2)
            if x>=skybbox[0] and x<=skybbox[1]-1 and y>=skybbox[2] and y<=skybbox[3]-1:
                #if x>=fbbox[0] and x<=fbbox[1]-1 and y>=fbbox[2] and y<=fbbox[3]-1:
                if r <= 1.0*skyradius and r >= 0.7*skyradius:
                    skymulti_index = (np.array([y]),np.array([x]))
                    skyravelindex[skycount] = utils.ravel_multi_index(skymulti_index,imshape)[0]
                    skycount += 1
                if r <= 1.0*hpsfnpix:
                    fmulti_index = (np.array([y]),np.array([x]))
                    fravelindex[fcount] = utils.ravel_multi_index(fmulti_index,imshape)[0]
                    fcount += 1
                if r <= fitradius:
                    multi_index = (np.array([y]),np.array([x]))
                    ravelindex[count] = utils.ravel_multi_index(multi_index,imshape)[0]
                    count += 1
    return (fravelindex,fcount,ravelindex,count,skyravelindex,skycount)

@njit
@cc.export('collatestarsinfo', '(UniTuple(i8,2),b1[:,:],f8[:],f8[:],i4,f8,f8)')
def collatestarsinfo(imshape,mask,starx,stary,hpsfnpix,fitradius,skyradius):
    """ Get full footprint, fitted pixels, and sky pixels data for all stars."""
    nstars = len(starx)
    nfpix = 2*hpsfnpix+1
    npix = int(np.floor(2*fitradius))+2
    # Full footprint arrays
    fravelindex = np.zeros((nstars,nfpix*nfpix),np.int64)
    fndata = np.zeros(nstars,np.int32)
    # Fitting pixel arrays
    ravelindex = np.zeros((nstars,npix*npix),np.int64)
    ndata = np.zeros(nstars,np.int32)
    skyravelindex = np.zeros((nstars,int(3.14*skyradius**2)),np.int64)
    skyndata = np.zeros(nstars,np.int32)
    for i in range(nstars):
        out = getstarinfo(imshape,mask,starx[i],stary[i],hpsfnpix,fitradius,skyradius)
        # full footprint information
        fravelindex1,fn1,ravelindex1,n1,skyravelindex1,skyn1 = out
        fravelindex[i,:] = fravelindex1
        fndata[i] = fn1
        # fitting pixel information
        ravelindex[i,:] = ravelindex1
        ndata[i] = n1
        # sky pixel information
        skyravelindex[i,:] = skyravelindex1
        skyndata[i] = skyn1
    # Trim arrays
    maxfn = np.max(fndata)
    fravelindex = fravelindex[:,:maxfn]
    maxn = np.max(ndata)
    ravelindex = ravelindex[:,:maxn]
    maxskyn = np.max(skyndata)
    skyravelindex = skyravelindex[:,:maxskyn]
    return (fravelindex,fndata,ravelindex,ndata,skyravelindex,skyndata)


@njit
@cc.export('initstararrays', '(f8[:,:],f8[:,:],b1[:,:],f8[:,:],i8,f8,f8,b1)')
def initstararrays(image,error,mask,tab,psfnpix,fitradius,skyradius,skyfit):
    """ Initialize all of the star arrays."""

    # Star arrays
    #------------
    #  -full footprint: pixels of a star within the psf radius and not masked
    #  -fitting pixels: pixels of a star within its fitting radius and not masked
    #  -sky pixels: pixels in an annulus around a star and not masked
    #  -flat pixels: all fitting pixels of all stars combined (unique),
    #                  these are the pixels that we are actually fitting
    # Full footprint information
    #   starravelindex -
    #   starndata - number of pixels for star each star
    # Fitting pixel information
    #   starfitravelindex - fitting pixel index into the raveled 1D full image/resid array
    #   starfitndata - number of fitting pixels for each star
    # Sky pixel information
    #   skyravelindex - sky pixel index into the raveled 1D full image/resid array
    #   skyndata - number of star pixels for each star
    # Flat information
    #   xflat/yflat - x/y values of the flat pixels
    #   indflat - index of the flat pixels into the full 1D raveled arrays
    #   starflat_index - index of all flat pixels that are within a star's full footprint
    #                         index into the flat 1D array
    #   starflat_ndata - number of flat pixels for each star
    
    nstars = len(tab)
    ny,nx = image.shape                             # save image dimensions, python images are (Y,X)
    #imshape = np.array([ny,nx])
    imshape = (ny,nx)
    im = image.copy().astype(np.float64)
    err = error.copy().astype(np.float64)
    msk = mask.copy().astype(np.bool_)
    xx,yy = utils.meshgrid(np.arange(nx),np.arange(ny))
    xx = xx.flatten()
    yy = yy.flatten()
    # Order stars by flux, brightest first
    si = np.argsort(tab[:,1])[::-1]       # largest amp first
    startab = tab[si]                     # ID, amp, xcen, ycen
    initpars = startab[:,1:4]             # amp, xcen, ycen
    # Initialize the parameter array
    initpars = startab[:,1:4]
    npars = nstars*3
    if skyfit:
        npars += 1
    pars = np.zeros(npars,float) # amp, xcen, ycen
    pars[0:3*nstars:3] = initpars[:,0]
    pars[1:3*nstars:3] = initpars[:,1]
    pars[2:3*nstars:3] = initpars[:,2]

    # Get information for all the stars
    xcen = pars[1:3*nstars:3]
    ycen = pars[2:3*nstars:3]
    hpsfnpix = psfnpix//2
    out = collatestarsinfo(imshape,msk,xcen,ycen,hpsfnpix,fitradius,skyradius)
    starravelindex,starndata,starfitravelindex,starfitndata,skyravelindex,skyndata = out

    # Put indices of all the unique fitted pixels into one array
    ntotpix = np.sum(starfitndata)
    allravelindex = np.zeros(ntotpix,np.int64)
    count = 0
    for i in range(nstars):
        n1 = starfitndata[i]
        ravelind1 = starfitravelindex[i,:n1]
        allravelindex[count:count+n1] = ravelind1
        count += n1
    allravelindex = np.unique(allravelindex)
    ravelindex = allravelindex
    ntotpix = len(allravelindex)

    # Combine all of the X and Y values (of the pixels we are fitting) into one array
    ntotpix = np.sum(starfitndata)
    xall = np.zeros(ntotpix,np.int32)
    yall = np.zeros(ntotpix,np.int32)
    count = 0
    for i in range(nstars):
        n1 = starfitndata[i]
        ravelindex1 = starfitravelindex[i,:n1]
        xdata1 = xx[ravelindex1]
        ydata1 = yy[ravelindex1]
        xall[count:count+n1] = xdata1
        yall[count:count+n1] = ydata1
        count += n1
    
    # Create 1D unraveled indices, python images are (Y,X)
    ind1 = utils.ravel_multi_index((yall,xall),imshape)
    # Get unique indices and inverse indices
    #   the inverse index list takes you from the full/duplicated pixels
    #   to the unique ones
    uind1,uindex1,invindex = utils.unique_index(ind1)
    ntotpix = len(uind1)
    ucoords = utils.unravel_index(uind1,image.shape)
    yflat = ucoords[:,0]
    xflat = ucoords[:,1]
    # x/y coordinates of the unique fitted pixels
    indflat = uind1
    
    # Save information on the "flattened" and unique fitted pixel arrays
    imflat = np.zeros(len(uind1),np.float64)
    imflat[:] = image.ravel()[uind1]
    errflat = np.zeros(len(uind1),np.float64)
    errflat[:] = error.ravel()[uind1]
    resflat = imflat.copy()
    
    # Add inverse index for the fitted pixels
    #  to be used with imflat/resflat/errflat/xflat/yflat/indflat
    maxfitpix = np.max(starfitndata)
    starfitinvindex = np.zeros((nstars,maxfitpix),np.int64)-1
    for i in range(nstars):
        n1 = starfitndata[i]
        if i==0:
            invlo = 0
        else:
            invlo = np.sum(starfitndata[:i])
        invindex1 = invindex[invlo:invlo+n1]
        starfitinvindex[i,:n1] = invindex1
    
    # For all of the fitting pixels, find the ones that
    # a given star contributes to (basically within its PSF radius)
    starflat_index = np.zeros((nstars,ntotpix),np.int64)-1
    starflat_ndata = np.zeros(nstars,np.int64)
    for i in range(ntotpix):
        x1 = xflat[i]
        y1 = yflat[i]
        for j in range(nstars):
            pars1 = pars[3*i:3*i+3]
            r = np.sqrt((xflat[i]-pars[1])**2 + (yflat[i]-pars[2])**2)
            if r <= psfnpix:
                starflat_index[j,starflat_ndata[j]] = i
                starflat_ndata[j] += 1
    maxndata = np.max(starflat_ndata)
    starflat_index = starflat_index[:,:maxndata]

    return (im,err,msk,xx,yy,pars,npars,
            starravelindex,starndata,starfitravelindex,starfitndata,skyravelindex,skyndata,
            xflat,yflat,indflat,imflat,errflat,resflat,ntotpix,
            starfitinvindex,starflat_index,starflat_ndata)


# @njit
# @cc.export('sky', '(f8[:,:],f8[:,:],unicode_type,f8,f8)')
# def sky(image,modelim,method='sep',rin=None,rout=None):
#     """ (Re)calculate the sky."""
#     # Remove the current best-fit model
#     resid = image-modelim    # remove model
#     skyim = utils.sky(resid)
#     # Calculate sky value for each star
#     #  use center position
#     for i in range(nstars):
#         starsky[i] = skyim[int(np.round(starycen[i])),
#                            int(np.round(starxcen[i]))]
#     return starsky
        

#     #     # Annulus aperture
#     #     elif method=='annulus':
#     #         if rin is None:
#     #             rin = self.psf.fwhm()*1.5
#     #         if rout is None:
#     #             rout = self.psf.fwhm()*2.5
#     #         positions = list(zip(self.starxcen,self.starycen))
#     #         annulus = CircularAnnulus(positions,r_in=rin,r_out=rout)
#     #         for i in range(self.nstars):
#     #             annulus_mask = annulus[i].to_mask(method='center')
#     #             annulus_data = annulus_mask.multiply(resid,fill_value=np.nan)
#     #             data = annulus_data[(annulus_mask.data>0) & np.isfinite(annulus_data)]
#     #             mean_sigclip, median_sigclip, _ = sigma_clipped_stats(data,stdfunc=dln.mad)
#     #             self.starsky[i] = mean_sigclip
#     #         if hasattr(self,'skyim') is False:
#     #             self.skyim = np.zeros(self.image.shape,float)
#     #         if self.skyim is None:
#     #             self.skyim = np.zeros(self.image.shape,float)
#     #         self.skyim += np.median(self.starsky)
#     #     else:
#     #         raise ValueError("Sky method "+method+" not supported")

@njit
@cc.export('psf', '('+xdatatype+',f8[:],'+psfdatatype+')')
def psf(xdata,pars,psfdata):
    """ Thin wrapper for getting a PSF model for a single star."""
    xind,yind = xdata
    psftype,psfparams,psflookup,psforder,imshape = psfdata
    im1,_ = mnb.psf(xind,yind,pars,psftype,psfparams,psflookup,
                    imshape,False,False)
    return im1
        

@njit
@cc.export('psfjac', '('+xdatatype+',f8[:],'+psfdatatype+')')
def psfjac(xdata,pars,psfdata):
    """ Thin wrapper for getting the PSF model and Jacobian for a single star."""
    xind,yind = xdata
    psftype,psfparams,psflookup,psforder,imshape = psfdata
    im1,jac1 = mnb.psf(xind,yind,pars,psftype,psfparams,psflookup,
                       imshape,True,False)
    return im1,jac1
    

@njit
@cc.export('model', '('+psfdatatype+','+freezedatatype+','+flatdatatype+',f8[:],b1,b1,b1)')
def model(psfdata,freezedata,flatdata,pars,trim=False,allparams=False,verbose=False):
    """ Calculate the model for the stars and pixels we are fitting."""
    
    if verbose==True:
        print('model: ',pars)

    # Unpack the PSF data
    psftype,psfparams,psflookup,psforder,imshape = psfdata
    # Unpack freeze information
    freezepars,freezestars = freezedata
    # Unpack flat information
    starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix = flatdata
    
    # Args are [amp,xcen,ycen] for all Nstars + sky offset
    # so 3*Nstars+1 parameters
    
    nstars = len(freezestars)
    npars = len(pars)
    nfreezestars = np.sum(freezestars)
    nfreezepars = np.sum(freezepars)
    nfreepars = npars-nfreezepars
    skyfit = (npars % 3 != 0)
    
    # Figure out the parameters of ALL the stars
    #  some stars and parameters are FROZEN
    # if nfreezepars>0 and allparams==False:
    #     allpars = np.zeros(len(freezepars),np.float64)
    #     allpars[np.where(freezepars==False)] = pars
    # else:
    #     allpars = pars
    allpars = pars

    allim = np.zeros(ntotpix,np.float64)
    
    # Loop over the stars and generate the model image        
    # ONLY LOOP OVER UNFROZEN STARS
    dostars = np.arange(nstars)[np.where(freezestars==False)]
    for i in dostars:
        pars1 = allpars[i*3:(i+1)*3]
        n1 = starflat_ndata[i]
        invind1 = starflat_index[i,:n1]
        xind1 = xflat[invind1]
        yind1 = yflat[invind1]
        xdata1 = (xind1,yind1)
        # we need the inverse index to the unique fitted pixels
        im1 = psf(xdata1,pars1,psfdata)
        allim[invind1] += im1
        
    if skyfit:
        allim += allpars[-1]  # add sky offset
        
    # if trim and nusepix<self.ntotpix:            
    #     unused = np.arange(self.ntotpix)[~usepix]
    #     allim = np.delete(allim,unused)
        
    return allim


@njit
@cc.export('fullmodel', '('+psfdatatype+','+stardatatype+',f8[:])')
def fullmodel(psfdata,stardata,pars):
    """ Calculate the model for all the stars and the full footprint."""
    
    # Unpack the PSF data
    psftype,psfparams,psflookup,psforder,imshape = psfdata
    # Unpack flat information
    starravelindex,starndata,xx,yy = stardata

    nstars = len(starndata)
    npars = len(pars)
    im = np.zeros(imshape[0]*imshape[1],np.float64)
    
    # Loop over all stars and generate the model image        
    for i in range(nstars):
        pars1 = pars[i*3:(i+1)*3]
        n1 = starndata[i]
        ravelindex1 = starravelindex[i,:n1]
        xind1 = xx[ravelindex1]
        yind1 = yy[ravelindex1]
        xdata1 = (xind1,yind1)
        # we need the inverse index to the unique fitted pixels
        im1 = psf(xdata1,pars1,psfdata)
        im[ravelindex1] += im1
        
    return im


@njit
@cc.export('jac', '('+psfdatatype+','+freezedatatype+','+flatdatatype+',f8[:],b1,b1)')
def jac(psfdata,freezedata,flatdata,pars,trim=False,allparams=False):
    """ Calculate the jacobian for the pixels and parameters we are fitting"""

    # Unpack the PSF data
    psftype,psfparams,psflookup,psforder,imshape = psfdata
    # Unpack freeze information
    freezepars,freezestars = freezedata
    # Unpack flat information
    starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix = flatdata
    
    # Args are [amp,xcen,ycen] for all Nstars + sky offset
    # so 3*Nstars+1 parameters

    nstars = len(freezestars)
    npars = len(pars)
    nfreezestars = np.sum(freezestars)
    nfreezepars = np.sum(freezepars)
    nfreepars = npars-nfreezepars
    skyfit = (npars % 3 != 0)

    # Figure out the parameters of ALL the stars
    #  some stars and parameters are FROZEN
    # if nfreezepars>0 and allparams==False:
    #     allpars = np.zeros(len(freezepars),np.float64)
    #     if len(pars) != (len(freezepars)-nfreezepars):
    #         print('problem')
    #         print('len(pars):',len(pars))
    #         print('len(freezepars):',len(freezepars))
    #         print('nfreezepars:',nfreezepars)
    #         return np.zeros(1,np.float64)+np.nan,np.zeros((1,1),np.float64)+np.nan
    #     allpars[np.where(freezepars==False)] = pars
    # else:
    #     allpars = pars
    allpars = pars
        
    im = np.zeros(ntotpix,np.float64)
    jac = np.zeros((ntotpix,len(pars)),np.float64)    # image covered by star
    usepix = np.zeros(ntotpix,np.bool_)
    
    # Loop over the stars and generate the model image        
    # ONLY LOOP OVER UNFROZEN STARS
    dostars = np.arange(nstars)[np.where(freezestars==False)]
    for i in dostars:
        pars1 = allpars[i*3:(i+1)*3]
        n1 = starflat_ndata[i]
        invind1 = starflat_index[i,:n1]
        xind1 = xflat[invind1]
        yind1 = yflat[invind1]
        xdata1 = (xind1,yind1)
        # we need the inverse index to the unique fitted pixels
        im1,jac1 = psfjac(xdata1,pars1,psfdata)
        jac[invind1,i*3] = jac1[:,0]
        jac[invind1,i*3+1] = jac1[:,1]
        jac[invind1,i*3+2] = jac1[:,2]
        im[invind1] += im1
        usepix[invind1] = True
        
    # Sky gradient
    if skyfit:
        jac[:,-1] = 1
            
    # Remove frozen columns
    #if nfreezepars>0 and allparams==False:
    if nfreezepars>0:
        origjac = jac.copy()
        jac = np.zeros((ntotpix,nfreepars),np.float64)
        freeparind, = np.where(freezepars==False)
        for count,i in enumerate(freeparind):
            jac[:,count] = origjac[:,i]

    # # Trim out unused pixels
    # if trim and nusepix<ntotpix:
    #     unused = np.arange(ntotpix)[~usepix]
    #     origjac = jac
    #     jac = np.zeros((nusepix,nfreepars),np.float64)
    #     usepixind, = np.where(usepix==True)
    #     im = im[usepixind]
    #     for count,i in enumerate(usepixind):
    #         for j in range(nfreepars):
    #             jac[count,j] = origjac[i,j]
        
    return im,jac


@njit
@cc.export('chisqflat', '('+freezedatatype+','+flatdatatype+','+psfdatatype+',f8[:],f8[:],f8[:])')
def chisqflat(freezedata,flatdata,psfdata,resflat,errflat,pars):
    """ Return chi-squared of the flat data"""
    # Note this ignores any frozen stars
    bestmodel = model(psfdata,freezedata,flatdata,pars,False,True)   # trim, allparams
    chisq = np.sum(resflat**2/errflat**2)
    return chisq


@njit
@cc.export('cov', '('+psfdatatype+','+freezedatatype+','+covflatdatatype+',f8[:])')
def cov(psfdata,freezedata,covflatdata,pars):
    """ Determine the covariance matrix."""
    
    # Unpack the PSF data
    psftype,psfparams,psflookup,psforder,imshape = psfdata
    # Unpack freeze information
    freezepars,freezestars = freezedata
    # Unpack flat information
    starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix,imflat,errflat,skyflat = covflatdata
    
    # https://stats.stackexchange.com/questions/93316/
    #             parameter-uncertainty-after-non-linear-least-squares-estimation
    # more background here, too: http://ceres-solver.org/nnls_covariance.html        
    # Hessian = J.T * T, Hessian Matrix
    #  higher order terms are assumed to be small
    # https://www8.cs.umu.se/kurser/5DA001/HT07/lectures/lsq-handouts.pdf
    flatdata2 = (starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix)
    bestmodel,mjac = jac(psfdata,freezedata,flatdata2,pars,False,True)   # trim, allparams
    # Weights
    #   If weighted least-squares then
    #   J.T * W * J
    #   where W = I/sig_i**2
    wt = np.diag(1/errflat**2)
    hess = mjac.T @ (wt @ mjac)
    # cov = H-1, covariance matrix is inverse of Hessian matrix
    cov_orig = utils.inverse(hess)
    # Rescale to get an unbiased estimate
    # cov_scaled = cov * (RSS/(m-n)), where m=number of measurements, n=number of parameters
    # RSS = residual sum of squares
    #  using rss gives values consistent with what curve_fit returns
    resid = imflat-bestmodel-skyflat
    # Use chi-squared, since we are doing the weighted least-squares and weighted Hessian
    chisq = np.sum(resid**2/errflat**2)        
    cov = cov_orig * (chisq/(ntotpix-len(pars)))  # what MPFITFUN suggests, but very small
        
    return cov


@njit
@cc.export('dofreeze', '(b1[:],f8[:],'+freezedatatype+','+flatdatatype+','+psfdatatype+',f8[:,:],f8[:])')
def dofreeze(frzpars,pars,freezedata,flatdata,psfdata,resid,resflat):
    """ Freeze par/stars."""
    # frzpars: boolean array for all FREE parameters specifying which ones should be frozen [Nfreepars]
    # pars: all parameters  [Npars]
    
    freezepars,freezestars = freezedata
    starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix = flatdata
    npars = len(freezepars)
    nstars = len(freezestars)

    # freezepars: boolean array for all parameters [3*nstars], True means frozen
    # freezestars: boolean array for all stars [nstars], True means frozen
    
    # Freeze parameters/stars that converged
    #  also subtract models of fixed stars
    #  also return new free parameters
    freeparsind, = np.where(freezepars==False)      # parameters for free stars
    frozenparsind, = np.where(freezepars==True)     # parameters for frozen stars
    # Save the original freeze values
    origfreezepars = freezepars.copy()
    origfreezestars = freezestars.copy()
    # Update freeze values for current "free" parameters
    freezepars[freeparsind] = frzpars            # stick in updated frozen values for "free" parameters
    # Only freeze full stars, not individual parameters (either all 3 or none)
    #  reconstruct freezestars from the updated freezepars array
    freezestars = np.sum(freezepars[0:3*nstars].copy().reshape(nstars,3),axis=1)==3
    #  now make sure only parameters are frozen for frozen stars. again, all or nothing
    if np.sum(freezestars)>0:
        #allfrozenstarsind, = np.where(freezestars==True)
        for i in np.where(freezestars==True)[0]:
            freezepars[i*3:i*3+3] = True
    # Subtract model for newly frozen stars
    newfreezestars, = np.where((origfreezestars==False) & (freezestars==True))
    
    # Freezing more stars
    if len(newfreezestars)>0:
        for i in range(len(newfreezestars)):
            istar = newfreezestars[i]
            freezestars[istar] = True
            freezepars[3*istar:3*istar+3] = True
            # Subtract model of frozen star from residuals
            pars1 = pars[3*istar:3*istar+3]
            n1 = starflat_ndata[istar]
            invind1 = starflat_index[istar,:n1]
            xind1 = xflat[invind1]
            yind1 = yflat[invind1]
            xdata1 = (xind1,yind1)
            im1 = psf(xdata1,pars1,psfdata)
            resflat[invind1] -= im1
            resid[indflat[invind1]] -= im1
            
        # Get the new array of free parameters
        freeparsind, = np.where(freezepars==False)
        bestpar = pars[freeparsind]
        nfreezepars = np.sum(freezepars)
        nfreezestars = np.sum(freezestars)
        
    return freezepars,freezestars,resid,resflat


@njit
@cc.export('groupfit', '(i8,f8[:],i8,f8[:,:,:],f8,f8[:,:],f8[:,:],b1[:,:],f8[:,:],f8,i8,f8,i8,b1,b1,b1)')
def groupfit(psftype,psfparams,psfnpix,psflookup,psfflux,
             image,error,mask,tab,fitradius,maxiter=10,
             minpercdiff=0.5,reskyiter=2,nofreeze=False,
             skyfit=False,verbose=False):
    """
    Fit PSF to a group of stars simultaneously.

    Parameters
    ----------
    psftype : int
       PSF type.
    psfparams: numpy array
       PSF analytical parameters.
    psfnpix : int
       Number of pixels in the PSF footprint.
    psflookup : numpy array
       PSF lookup table.  Must be 3D.
    psfflux : float
       PSF flux.
    image : numpy array
       The image to fit.
    error : numpy array
       Uncertainty for image.
    mask : numpy array
       Boolean mask for image.  True - bad, False - good.
    tab : numpy array
       Table of initial guesses for each star.
       id, amp, xcen, ycen
    maxiter : int, optional
       Maximum number of iterations to allow.  Only for methods "cholesky", "qr" or "svd".
       Default is 10.
    minpercdiff : float, optional
       Minimum percent change in the parameters to allow until the solution is
       considered converged and the iteration loop is stopped.  Only for methods
       "cholesky", "qr" and "svd".  Default is 0.5.
    reskyiter : int, optional
       After how many iterations to re-calculate the sky background. Default is 2.
    nofreeze : boolean, optional
       Do not freeze any parameters even if they have converged.  Default is False.
    skyfit : boolean, optional
       Fit a constant sky offset with the stellar parameters.  Default is False.
    verbose : boolean, optional
       Verbose output.

    Returns
    -------
    out : table
       Table of best-fitting parameters for each star.
       id, amp, amp_error, x, x_err, y, y_err, sky
    model : numpy array
       Best-fitting model of the stars and sky background.
    sky : numpy array
       Best-fitting sky image.

    Examples
    --------

    out,model,sky = groupfit(psf,image,error,mask,tab)
    
    """

    #start = clock()
    
    nstars = len(tab)
    skyradius = psfnpix//2 + 10
    ny,nx = image.shape                             # save image dimensions, python images are (Y,X)
    #imshape = np.array([ny,nx])
    imshape = (ny,nx)
    
    print('groupfit() 1')
    
    # PSF information
    #----------------
    _,_,npsforder = psflookup.shape
    psforder = 0
    if npsforder==4:
        psforder = 1
    if psflookup.ndim != 3:
        raise Exception('psflookup must have 3 dimensions')
    psforder = psflookup.shape[2]
    if psforder>1:
        psforder = 1
    # Package up the PSF information into a tuple to pass to the functions
    psfdata = (psftype,psfparams,psflookup,psforder,imshape)

    print('groupfit() 2')

    #return psfdata
    
    # Star arrays
    #------------
    #  -full footprint: pixels of a star within the psf radius and not masked
    #  -fitting pixels: pixels of a star within its fitting radius and not masked
    #  -sky pixels: pixels in an annulus around a star and not masked
    #  -flat pixels: all fitting pixels of all stars combined (unique),
    #                  these are the pixels that we are actually fitting
    # Full footprint information
    #   starravelindex -
    #   starndata - number of pixels for star each star
    # Fitting pixel information
    #   starfitravelindex - fitting pixel index into the raveled 1D full image/resid array
    #   starfitndata - number of fitting pixels for each star
    # Sky pixel information
    #   skyravelindex - sky pixel index into the raveled 1D full image/resid array
    #   skyndata - number of star pixels for each star
    # Flat information
    #   xflat/yflat - x/y values of the flat pixels
    #   indflat - index of the flat pixels into the full 1D raveled arrays
    #   starflat_index - index of all flat pixels that are within a star's full footprint
    #                         index into the flat 1D array
    #   starflat_ndata - number of flat pixels for each star
    
    # Initialize the star arrays
    initdata = initstararrays(image,error,mask,tab,psfnpix,fitradius,skyradius,skyfit)
    im,err,msk,xx,yy,pars,npars = initdata[:7]
    starravelindex,starndata,starfitravelindex,starfitndata,skyravelindex,skyndata = initdata[7:13]
    xflat,yflat,indflat,imflat,errflat,resflat,ntotpix = initdata[13:20]
    starfitinvindex,starflat_index,starflat_ndata = initdata[20:]
    
    # Package up the star information into tuples for easy transport
    stardata = (starravelindex,starndata,xx,yy)
    flatdata = (starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix)

    print('groupfit() 3')
    
    # Image arrays
    #-------------
    # Full arrays:
    #   im/err/mask - full 1D raveled image, error and mask arrays
    #   resid - full 1D raveled residual image with the smooth sky subtracted
    #             AND any frozen stars subtracted
    # Flat arrays:
    #   indflat - index of the flat pixels into the full 1D raveled arrays
    #   imflat/errflat - 1D raveled image and error arrays for the flat/fitting pixels
    #   skyflat - smooth sky values for the flat pixels
    #   resflat - flat pixel values of "resid"
    
    # Create initial sky image
    #  improve initial sky estimate by removing initial models
    tresid = np.zeros(imshape[0]*imshape[1],np.float64)
    tresid[:] = im.copy().astype(np.float64).flatten()
    for i in range(nstars):
        pars1 = pars[3*i:3*i+3]
        n1 = starndata[i]
        ravelind1 = starravelindex[i]
        xind1 = xx[ravelind1]
        yind1 = yy[ravelind1]
        xdata1 = (xind1,yind1)
        m = psf(xdata1,pars1,psfdata)
        tresid[ravelind1] -= m
    skyim = utils.sky(tresid.copy().reshape(imshape[0],imshape[1])).flatten()
    skyflat = skyim[indflat]

    print('groupfit() 4')
    
    # Initialize RESID and subtract initial smooth sky
    resid = np.zeros(imshape[0]*imshape[1],np.float64)
    resid[:] = im.copy().astype(np.float64).flatten()   # flatten makes it easier to modify
    resid[:] -= skyim          # subtract smooth sky
    resflat = resid[indflat]   # initialize resflat

    
    # Perform the fitting
    #--------------------

    # Initialize arrays for the stars and freezing parameters
    starsky = np.zeros(nstars,np.float64)
    starniter = np.zeros(nstars,np.int64)
    starchisq = np.zeros(nstars,np.float64)
    starrms = np.zeros(nstars,np.float64)
    freezestars = np.zeros(nstars,np.bool_)
    freezepars = np.zeros(len(pars),np.bool_)
    # freezepars: boolean array for all parameters [3*nstars], True means frozen
    # freezestars: boolean array for all stars [nstars], True means frozen
    
    print('groupfit() 5')
    
    # Initial estimates
    initpar = pars.copy()
    # Make bounds
    #  this requires all 3*Nstars parameters to be input
    bounds = utils.mkbounds(initpar,imshape,2)    
    
    # Iterate
    bestpar_all = initpar.copy()
    allpars = initpar.copy()
    bestpar = initpar.copy()
    # centroids fixed, only keep amps
    #if recenter==False:
    #    initpar = initpar[0::3]  # amps and sky offset
    # sky offset fixed
    #if skyfit==False:
    #    initpar = initpar[0:-1]  # no sky offset
    niter = 1
    maxpercdiff = 1e10
    maxsteps = utils.steps(initpar)  # maximum steps, requires all 3*Nstars parameters
    nfreepars = len(pars)
    nfreestars = nstars
    # While loop
    #   keep looping until: (a) reached max iterations, (b) changes are tiny
    #                      or (c) there are no free stars left
    while (niter<maxiter and maxpercdiff>minpercdiff and nfreestars>0):
        #start0 = clock()

        # bestpar: current best values for the free parameters
        # pars: current best values for ALL parameters
        
        # --- Jacobian solvers ---
        # Get the Jacobian and model
        #  only for pixels that are affected by the "free" parameters
        freezedata = (freezepars,freezestars)
        model0,j = jac(psfdata,freezedata,flatdata,pars,False,True)  # trim, allparams
        # Residuals
        skyflat = skyim[indflat]
        dy = resflat-skyflat-model0
        # Weights
        wt = 1/errflat**2
        # Solve linear least squares with the Jacobian
        dbeta_free = utils.qr_jac_solve(j,dy,weight=wt)
        dbeta_free[~np.isfinite(dbeta_free)] = 0.0  # deal with NaNs, shouldn't happen
        dbeta = np.zeros(len(pars),np.float64)
        dbeta[np.where(freezepars==False)] = dbeta_free

        print('groupfit() 6')
        
        # --- Perform line search ---
        #  move the solution along the dbeta vector to find the lowest chisq
        #  get models: pars + 0*dbeta, pars + 0.5*dbeta, pars + 1.0*dbeta
        freezedata = (freezepars,freezestars)
        model1 = model(psfdata,freezedata,flatdata,pars+0.5*dbeta,False,True)   # trim, allparams
        model2 = model(psfdata,freezedata,flatdata,pars+dbeta,False,True)       # trim, allparams
        chisq0 = np.sum((resflat-model0)**2/errflat**2)
        chisq1 = np.sum((resflat-model1)**2/errflat**2)
        chisq2 = np.sum((resflat-model2)**2/errflat**2)
        if verbose:
            print('linesearch:',chisq0,chisq1,chisq2)
        # Use quadratic bisector on the three points to find the lowest chisq 
        alpha = utils.quadratic_bisector(np.array([0.0,0.5,1.0]),
                                         np.array([chisq0,chisq1,chisq2]))
        # The bisector can be negative (outside of the acceptable range) if the chisq shape is concave
        if alpha <= 0 and np.min(np.array([chisq0,chisq1,chisq2]))==chisq2:
            alpha = 1.0
        #alpha = np.minimum(np.maximum(alpha,0.0),1.0)  # 0<alpha<1
        alpha = utils.clip(alpha,0.0,1.0)  # 0<alpha<1
        if np.isfinite(alpha)==False:
            alpha = 1.0
        pars_new = pars + alpha * dbeta
        new_dbeta_free = alpha * dbeta_free
        new_dbeta = np.zeros(len(pars),float)
        new_dbeta[np.where(freezepars==False)] = new_dbeta_free

        print('groupfit() 7')
        
        # Update the free parameters and impose step limits and bounds
        maxsteps = utils.steps(pars)  # maximum steps, requires all 3*Nstars parameters
        bounds = utils.mkbounds(pars,imshape,2)   
        oldpar = bestpar.copy()
        oldpar_all = pars.copy()
        bestpar_all = utils.newpars(pars,new_dbeta,bounds,maxsteps)            
        bestpar = bestpar_all[np.where(freezepars==False)]
        pars[np.where(freezepars==False)] = bestpar
        
        # Check differences and percent changes
        diff_all = np.abs(bestpar_all-oldpar_all)
        percdiff_all = diff_all.copy()*0
        percdiff_all[0:3*nstars:3] = diff_all[0:3*nstars:3]/np.maximum(oldpar_all[0:3*nstars:3],0.0001)*100  # amp
        percdiff_all[1:3*nstars:3] = diff_all[1:3*nstars:3]*100               # x
        percdiff_all[2:3*nstars:3] = diff_all[2:3*nstars:3]*100               # y
        if skyfit:
            percdiff_all[-1] = diff_all[-1]/np.maximum(np.abs(oldpar_all[-1]),1)*100
        diff = diff_all[np.where(freezepars==False)]
        percdiff = percdiff_all[np.where(freezepars==False)]
        
        # Freeze parameters/stars that converged
        #  this will subtract the model of converged stars from RESID and RESFLAT
        if nofreeze==False:
           frzpars = percdiff<=minpercdiff
           freezedata = (freezepars,freezestars)
           out = dofreeze(frzpars,pars,freezedata,flatdata,psfdata,resid,resflat)
           freezepars,freezestars,resid,resflat = out
        nfreestars = np.sum(freezestars==False)

        print('groupfit() 8')
        
        # Get model and chisq
        freezedata = (freezepars,freezestars)
        flatdata = (starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix)
        bestmodel = model(psfdata,freezedata,flatdata,pars,False,True)   # trim, allparams
        chisq = np.sum(resflat**2/errflat**2)
        
        if verbose:
            print('Iter = ',niter)
            print('Pars = ',pars)
            print('chisq = ',chisq)
            
        # Re-estimate the sky
        if niter % reskyiter == 0:
            if verbose:
                print('Re-estimating the sky')
            prevsky = skyim.copy()            
            # Get model of full footprints
            bestfullmodel = fullmodel(psfdata,stardata,pars)
            # Remove the current best-fit model and re-estimate the sky
            tresid = im-bestfullmodel.copy().reshape(imshape[0],imshape[1])
            skyim = utils.sky(tresid).flatten()
            # Update resid
            #  add previous sky back in
            # Remake the resid array using the new smooth sky
            resid[:] += prevsky
            resid[:] -= skyim   # subtract smooth sky
            resflat = resid[indflat]
            skyflat = skyim[indflat]
            
        #if verbose:
        #    print('iter dt =',(clock()-start0)/1e9,'sec.')
                
        niter += 1     # increment counter
        
    # Check that all starniter are set properly
    #  if we stopped "prematurely" then not all stars were frozen
    #  and didn't have starniter set
    starniter[np.where(starniter==0)] = niter

    print('groupfit() 9')
    
    # Unfreeze everything
    freezepars[:] = False
    freezestars[:] = False
    # Make final model
    finalmodel = fullmodel(psfdata,stardata,pars).copy().reshape(imshape[0],imshape[0])
    
    # Estimate uncertainties
    #   calculate covariance matrix
    freezedata = (freezepars,freezestars)
    covflatdata = (starflat_ndata,starflat_index,xflat,yflat,indflat,ntotpix,imflat,errflat,skyflat)
    cov1 = cov(psfdata,freezedata,covflatdata,pars)
    perror = np.sqrt(np.diag(cov1))
    
    if verbose:
        print('Best-fitting parameters: ',pars)
        print('Errors: ',perror)

    # # Calculate smooth sky value for each star
    # #  use center position
    # for i in range(nstars):
    #     x1 = int(np.round(starxcen[i]))
    #     y1 = int(np.round(starycen[i]))
    #     sind1 = utils.ravel_multi_index((y1,x1),imshape)
    #     starsky[i] = skyim[sind1]
    
    # Put in catalog
    outtab = np.zeros((nstars,15),np.float64)
    outtab[:,0] = np.arange(nstars)+1                # id
    outtab[:,1] = pars[0:3*nstars:3]                 # amp
    outtab[:,2] = perror[0:3*nstars:3]               # amp_error
    outtab[:,3] = pars[1:3*nstars:3]                 # x
    outtab[:,4] = perror[1:3*nstars:3]               # x_error
    outtab[:,5] = pars[2:3*nstars:3]                 # y
    outtab[:,6] = perror[2:3*nstars:3]               # y_error
    if skyfit:
        outtab[:,7] = starsky + pars[-1]             # sky
    else:
        outtab[:,7] = starsky                        # sky
    outtab[:,8] = outtab[:,1]*psfflux                # flux
    outtab[:,9] = outtab[:,2]*psfflux                # flux_error
    outtab[:,10] = -2.5*np.log10(np.maximum(outtab[:,8],1e-10))+25.0   # mag
    outtab[:,11] = (2.5/np.log(10))*outtab[:,9]/outtab[:,8]            # mag_error
    # rms
    # chisq
    outtab[:,14] = starniter                      # niter, what iteration it converged on
    
    # Recalculate chi-squared and RMS of fit for each star
    for i in range(nstars):
        n1 = starfitndata[i]
        ravelind1 = starfitravelindex[i,:n1]
        xind1 = xx[ravelind1]
        yind1 = yy[ravelind1]
        flux1 = image.ravel()[ravelind1].copy() #[yind1,xind1].copy()
        err1 = error.ravel()[ravelind1].copy()  #[yind1,xind1]
        pars1 = np.zeros(3,np.float64)
        pars1[0] = outtab[i,1]  # amp
        pars1[1] = outtab[i,3]  # x
        pars1[2] = outtab[i,5]  # y
        xdata1 = (xind1,yind1)
        model1 = psf(xdata1,pars1,psfdata)
        sky1 = outtab[i,7]
        chisq1 = np.sum((flux1-sky1-model1)**2/err1**2)/n1
        outtab[i,13] = chisq1
        # chi value, RMS of the residuals as a fraction of the amp
        rms1 = np.sqrt(np.mean(((flux1-sky1-model1)/pars1[0])**2))
        outtab[i,12] = rms1

    #if verbose:
    #    print('dt =',(clock()-start)/1e9,'sec.')
 
    return outtab,finalmodel,skyim

if __name__ == "__main__":
    cc.compile()



# fastarray.pyx
# cython: boundscheck=False, wraparound=False
import numpy as np       # Python-level for @, np.array
cimport numpy as cnp    # C-level for cdef ndarray

import cython

from libc.math cimport exp,sqrt,atan2,pi,NAN,log,log10,abs,pow,sin,cos

cdef class FastArrayND:
    # These are defined in the pxd file
    # cdef double[::1] data         # 1D contiguous memoryview
    # #cdef int[::1] shape           # shape as 1D contiguous memoryview
    # cdef int ndim
    # cdef Py_ssize_t size
    # cdef Py_ssize_t[:] shape
    # cdef Py_ssize_t[:] strides
    
    # -------------------
    # Constructor
    # -------------------
    def __init__(self, shape):        
        cdef int i, total=1

        #if not isinstance(shape, (tuple, list)):
        #    raise TypeError("shape must be a tuple or list")

        self.ndim = len(shape)
        if self.ndim < 1 or self.ndim > 3:
            raise ValueError("Only 1D, 2D, 3D supported")

        # store shape and strides as numpy intp arrays
        self.shape = np.array(shape, dtype=np.int64)
        self.strides = np.empty_like(self.shape)

        cdef Py_ssize_t stride = 1
        for i in range(self.ndim - 1, -1, -1):
            self.strides[i] = stride
            stride *= self.shape[i]

        # copy shape into contiguous memoryview
        self.shape = np.empty(self.ndim, dtype=np.int64)
        for i in range(self.ndim):
            self.shape[i] = shape[i]
            total *= shape[i]

        self.size = total
        self.data = np.zeros(total, dtype=np.float64)

    cdef inline Py_ssize_t _get_offset(self, tuple idx):
        """Convert N-D index into flat offset."""
        if len(idx) != self.ndim:
            raise IndexError(f"Expected {self.ndim} indices, got {len(idx)}")
        cdef Py_ssize_t offset = 0
        cdef int i
        for i in range(self.ndim):
            ii = idx[i]
            if ii < 0 or ii >= self.shape[i]:
                raise IndexError("index out of range")
            offset += ii * self.strides[i]
        return offset
        
    # -------------------
    # Internal binary operation (with broadcasting)
    # -------------------
    cdef void _binary_op(self, FastArrayND other, FastArrayND out, char op):
        cdef int i
        cdef int d0=1,d1=1,d2=1
        cdef int od0=1,od1=1,od2=1
        cdef int i0,i1,i2, idx_out, idx_self, idx_other
        if self.size == other.size:
            for i in range(self.size):
                if op=='+': out.data[i] = self.data[i] + other.data[i]
                elif op=='-': out.data[i] = self.data[i] - other.data[i]
                elif op=='*': out.data[i] = self.data[i] * other.data[i]
                elif op=='/': out.data[i] = self.data[i] / other.data[i]
                elif op=='^': out.data[i] = pow(self.data[i], other.data[i])
        elif other.size == 1:
            for i in range(self.size):
                if op=='+': out.data[i] = self.data[i] + other.data[0]
                elif op=='-': out.data[i] = self.data[i] - other.data[0]
                elif op=='*': out.data[i] = self.data[i] * other.data[0]
                elif op=='/': out.data[i] = self.data[i] / other.data[0]
                elif op=='^': out.data[i] = pow(self.data[i], other.data[0])
        else:
            # Simple broadcasting for 1D-3D arrays
            if self.ndim <= 3 and other.ndim <= 3:
                d0, d1, d2 = 1, 1, 1
                od0, od1, od2 = 1, 1, 1
                if self.ndim >=1: d0=self.shape[0]
                if self.ndim >=2: d1=self.shape[1]
                if self.ndim ==3: d2=self.shape[2]
                if other.ndim >=1: od0=other.shape[0]
                if other.ndim >=2: od1=other.shape[1]
                if other.ndim ==3: od2=other.shape[2]
                for i0 in range(d0):
                    for i1 in range(d1):
                        for i2 in range(d2):
                            idx_out = i0*d1*d2 + i1*d2 + i2
                            idx_self = idx_out
                            idx_other = (i0 if od0>1 else 0)*(od1*od2) + (i1 if od1>1 else 0)*od2 + (i2 if od2>1 else 0)
                            if op=='+': out.data[idx_out] = self.data[idx_self] + other.data[idx_other]
                            elif op=='-': out.data[idx_out] = self.data[idx_self] - other.data[idx_other]
                            elif op=='*': out.data[idx_out] = self.data[idx_self] * other.data[idx_other]
                            elif op=='/': out.data[idx_out] = self.data[idx_self] / other.data[idx_other]
                            elif op=='^': out.data[idx_out] = pow(self.data[idx_self], other.data[idx_other])
            else:
                raise ValueError("Unsupported shapes for broadcasting")

                

    # # -------------------
    # # Helper to handle scalars
    # # -------------------
    # cdef FastArrayND _ensure_array(self, other):
    #     if isinstance(other, FastArrayND):
    #         return other
    #     else:
    #         cdef FastArrayND tmp = FastArrayND(np.array([1], dtype=np.int32))
    #         tmp.data[0] = cython.cast(double, other)
    #         return tmp


    # cdef FastArrayND _binary_out(self, other, char op):
    #     cdef FastArrayND o = self._ensure_array(other)
    #     cdef FastArrayND out = FastArrayND(self.shape)
    #     self._binary_op(o, out, op)
    #     return out

    # -------------------
    # Helper to handle scalar or FastArrayND
    # -------------------
    cdef FastArrayND _ensure_array(self, other):
        """
        If other is FastArrayND, return as-is.
        If other is a Python scalar, wrap as FastArrayND of size 1.
        """
        cdef FastArrayND tmp
        if isinstance(other, FastArrayND):
            return other
        else:
            tmp = FastArrayND(np.array([1], dtype=np.int32))
            tmp.data[0] = cython.cast(double, other)
            return tmp

    # -------------------
    # Operator overloading with automatic scalar broadcasting
    # -------------------
    def __add__(self, other):
        cdef FastArrayND o = self._ensure_array(other)
        cdef FastArrayND out = FastArrayND(self.shape)
        self._binary_op(o, out, '+')
        return out

    def __sub__(self, other):
        cdef FastArrayND o = self._ensure_array(other)
        cdef FastArrayND out = FastArrayND(self.shape)
        self._binary_op(o, out, '-')
        return out

    def __mul__(self, other):
        cdef FastArrayND o = self._ensure_array(other)
        cdef FastArrayND out = FastArrayND(self.shape)
        self._binary_op(o, out, '*')
        return out

    def __truediv__(self, other):
        cdef FastArrayND o = self._ensure_array(other)
        cdef FastArrayND out = FastArrayND(self.shape)
        self._binary_op(o, out, '/')
        return out

    def __pow__(self, other):
        cdef FastArrayND o = self._ensure_array(other)
        cdef FastArrayND out = FastArrayND(self.shape)
        self._binary_op(o, out, '^')
        return out

    
    # -------------------
    # In-place operators
    # -------------------
    def __iadd__(self, other):
        self._binary_op(self._ensure_array(other), self, '+')
        return self
    def __isub__(self, other):
        self._binary_op(self._ensure_array(other), self, '-')
        return self
    def __imul__(self, other):
        self._binary_op(self._ensure_array(other), self, '*')
        return self
    def __itruediv__(self, other):
        self._binary_op(self._ensure_array(other), self, '/')
        return self
    def __ipow__(self, other):
        self._binary_op(self._ensure_array(other), self, '^')
        return self


    # -------------------
    # Elementwise ufuncs
    # -------------------
    cpdef FastArrayND exp(self):
        cdef FastArrayND out = FastArrayND(self.shape)
        cdef int i
        for i in range(self.size):
            out.data[i] = exp(self.data[i])
        return out
    cpdef FastArrayND log(self):
        cdef FastArrayND out = FastArrayND(self.shape)
        cdef int i
        for i in range(self.size):
            out.data[i] = log(self.data[i])
        return out
    cpdef FastArrayND sin(self):
        cdef FastArrayND out = FastArrayND(self.shape)
        cdef int i
        for i in range(self.size):
            out.data[i] = sin(self.data[i])
        return out
    cpdef FastArrayND cos(self):
        cdef FastArrayND out = FastArrayND(self.shape)
        cdef int i
        for i in range(self.size):
            out.data[i] = cos(self.data[i])
        return out

    # -------------------
    # Indexing / slicing
    # -------------------
    def __getitem__(self, idx):
        """
        Returns a FastArrayND slice or a scalar depending on idx.
        Supports:
            a[i]         -> 1D slice or scalar
            a[i,j]       -> 2D slice or scalar
            a[i:j, k:l]  -> subarray
        """
        cdef int ndim = self.ndim
        cdef FastArrayND out
        cdef int i, j, k, l
        cdef int shape0, shape1, shape2
        cdef int start, stop, step
        cdef int i_idx, j_idx, k_idx
        cdef int i_start, i_stop, i_step
        cdef int j_start, j_stop, j_step
        cdef int k_start, k_stop, k_step
        # Handle 1D array
        if ndim == 1:
            if isinstance(idx, int):
                return self.data[idx]
            elif isinstance(idx, slice):
                start, stop, step = idx.indices(self.shape[0])
                out = FastArrayND(np.array([stop-start], dtype=np.int32))
                for i in range(stop-start):
                    out.data[i] = self.data[start + i*step]
                return out
        # Handle 2D array
        elif ndim == 2:
            if isinstance(idx, tuple):
                i_idx, j_idx = idx
                # both integers -> scalar
                if isinstance(i_idx, int) and isinstance(j_idx, int):
                    return self.data[i_idx*self.shape[1]+j_idx]
                # slice -> subarray
                i_start, i_stop, i_step = i_idx.indices(self.shape[0]) if isinstance(i_idx, slice) else (i_idx,i_idx+1,1)
                j_start, j_stop, j_step = j_idx.indices(self.shape[1]) if isinstance(j_idx, slice) else (j_idx,j_idx+1,1)
                shape0 = (i_stop - i_start + i_step -1)//i_step
                shape1 = (j_stop - j_start + j_step -1)//j_step
                out = FastArrayND(np.array([shape0, shape1], dtype=np.int32))
                for i in range(shape0):
                    for j in range(shape1):
                        out.data[i*shape1 + j] = self.data[(i_start+i*i_step)*self.shape[1] + (j_start+j*j_step)]
                return out
        # Handle 3D array
        elif ndim == 3:
            if isinstance(idx, tuple):
                i_idx, j_idx, k_idx = idx
                # all integers -> scalar
                if isinstance(i_idx,int) and isinstance(j_idx,int) and isinstance(k_idx,int):
                    return self.data[i_idx*self.shape[1]*self.shape[2] + j_idx*self.shape[2] + k_idx]
                # slices -> subarray
                i_start,i_stop,i_step = i_idx.indices(self.shape[0]) if isinstance(i_idx,slice) else (i_idx,i_idx+1,1)
                j_start,j_stop,j_step = j_idx.indices(self.shape[1]) if isinstance(j_idx,slice) else (j_idx,j_idx+1,1)
                k_start,k_stop,k_step = k_idx.indices(self.shape[2]) if isinstance(k_idx,slice) else (k_idx,k_idx+1,1)
                shape0 = (i_stop-i_start+i_step-1)//i_step
                shape1 = (j_stop-j_start+j_step-1)//j_step
                shape2 = (k_stop-k_start+k_step-1)//k_step
                out = FastArrayND(np.array([shape0,shape1,shape2], dtype=np.int32))
                for i in range(shape0):
                    for j in range(shape1):
                        for k in range(shape2):
                            out.data[i*shape1*shape2 + j*shape2 + k] = \
                                self.data[(i_start+i*i_step)*self.shape[1]*self.shape[2] + 
                                          (j_start+j*j_step)*self.shape[2] + 
                                          (k_start+k*k_step)]
                return out
        raise IndexError("Unsupported indexing")

    # def __setitem__(self, Py_ssize_t i, double value):
    #     if i < 0 or i >= self.size:
    #         raise IndexError("index out of range")
    #     self.data[i] = value


    def __setitem__(self, idx, value):
        cdef double val = float(value)
        if self.ndim == 1:
            if not isinstance(idx, (int, np.integer)):
                raise IndexError("Expected integer index for 1D array")
            if idx < 0 or idx >= self.shape[0]:
                raise IndexError("index out of range")
            self.data[idx] = val
        elif isinstance(idx, tuple):
            self.data[self._get_offset(idx)] = val
        else:
            raise IndexError("Invalid index type")

    
    # def __setitem__(self, idx, double value):
    #     if self.ndim == 1:
    #         if not isinstance(idx, (int, np.integer)):
    #             raise IndexError("Expected integer index for 1D array")
    #         if idx < 0 or idx >= self.shape[0]:
    #             raise IndexError("index out of range")
    #         self.data[idx] = value
    #     elif isinstance(idx, tuple):
    #         self.data[self._get_offset(idx)] = value
    #     else:
    #         raise IndexError("Invalid index type")
        
    # # -------------------
    # # Matmul for 2D arrays
    # # -------------------
    # def matmul(self, FastArrayND other):
    #     if self.ndim != 2 or other.ndim != 2:
    #         raise ValueError("matmul only supports 2D arrays")
    #     cdef int m = self.shape[0]
    #     cdef int k1 = self.shape[1]
    #     cdef int k2 = other.shape[0]
    #     cdef int n = other.shape[1]
    #     if k1 != k2:
    #         raise ValueError("Incompatible shapes for matrix multiply")

    #     # Wrap buffers as Cython memoryviews for speed
    #     cdef cnp.ndarray[cnp.double_t, ndim=2, mode="c"] A = \
    #         np.ndarray((m, k1), dtype=np.float64, buffer=self.data)
    #     cdef cnp.ndarray[cnp.double_t, ndim=2, mode="c"] B = \
    #         np.ndarray((k2, n), dtype=np.float64, buffer=other.data)

    #     # Python-level matmul
    #     C = A @ B  # <-- must be Python-level

    #     # Copy to FastArrayND
    #     out = FastArrayND(np.array([m, n], dtype=np.int32))
    #     out.data[:] = C.ravel()
    #     return out

    # # -------------------
    # # Optional: operator overloading for @
    # # -------------------
    # def __matmul__(self, FastArrayND other):
    #     return self.matmul(other)

    # -------------------
    # Utility: get shape as tuple
    # -------------------
    cdef tuple shape_tuple(self):
        return tuple([self.shape[i] for i in range(self.ndim)])

    def to_numpy(self):
        """
        Return a NumPy array view of the FastArrayND data.
        """
        # cast data to a Cython memoryview for Python buffer interface
        cdef double[:] mv = self.data[:self.size]  # memoryview over contiguous buffer
        # convert to NumPy array (copies only if necessary)
        arr = np.array(mv, copy=False, dtype=np.float64)
        return arr.reshape(tuple(int(s) for s in self.shape))
    
    
    def flatten(self):
        """
        Return a new 1D FastArrayND copy of the data.
        Equivalent to numpy.flatten().
        """
        cdef FastArrayND out = FastArrayND((self.size,))
        cdef Py_ssize_t i
        for i in range(self.size):
            out.data[i] = self.data[i]
        return out

    def __repr__(self):
        """Return a string representation of the array."""
        # Convert to a NumPy array and use NumPy's printing
        return repr(self.to_numpy())



#!/usr/bin/env python

"""CCDDATA.PY - Thin wrapper around CCDData class

"""

__authors__ = 'David Nidever <dnidever@montana.edu?'
__version__ = '20210908'  # yyyymmdd

import sys
import time
import numpy as np
from astropy.nddata import CCDData as CCD,StdDevUncertainty
from astropy.wcs import WCS
from astropy.io import fits
from photutils.aperture import BoundingBox as BBox
from copy import deepcopy
from dlnpyutils import utils as dln
from . import sky as psky


def poissonnoise(data,gain=1.0,rdnoise=0.0):
    """ Generate Poisson noise model (ala DAOPHOT)."""
    # gain
    # rdnoise
    noise = np.sqrt(np.maximum(data,0)/gain + rdnoise**2)
    noise = np.maximum(noise,1)
    return noise

def getgain(image):
    """ Get the gain from the header."""

    gain = 1.0  # default
    
    # check if there's a header
    if hasattr(image,'meta'):
        if image.meta is not None:
            # Try all versions of gain
            for f in ['gain','egain','gaina']:
                hgain = image.meta.get(f)
                if hgain is not None:
                    gain = float(hgain)
                    break
    return gain
    
def getrdnoise(image):
    """ Get the read noise from the header."""

    rdnoise = 0.0  # default
    
    # check if there's a header
    if hasattr(image,'meta'):
        if image.meta is not None:
            # Try all versions of rdnoise
            for f in ['rdnoise','readnois','enoise','rdnoisea']:
                hrdnoise = image.meta.get(f)
                if hrdnoise is not None:
                    rdnoise = float(hrdnoise)
                    break
    return rdnoise

def mkbbox(data):
    """ Make BoundingBox and x,y arrays for data."""

    ndim = data.ndim
    if ndim==0:
        bbox = BoundingBox(0,0,0,0)
        x = None
        y = None         
    elif ndim==1:
        nx, = data.shape
        bbox = BoundingBox(0,nx,0,0)
        x = np.arange(bbox.xrange[0],bbox.xrange[-1])
        y = None
    elif ndim==2:
        ny,nx = data.shape
        bbox = BoundingBox(0,nx,0,ny)
        x = np.arange(bbox.xrange[0],bbox.xrange[-1])
        y = np.arange(bbox.yrange[0],bbox.yrange[-1])
    else:
        raise ValueError('3D CCDData not supported')
    return bbox,x,y


class CCDData(CCD):
    """
    
    A container for image data.  This is based on the astropy CCDdata class, but some
    functionality has been added.

    Parameters
    ----------
    data : `numpy.ndarray`-like or `NDData`-like
        The dataset.

    error : any type, optional
        Uncertainty in the dataset.  This must be a "reguar" standard deviation
        uncertainty. Defaults to ``None``.

    mask : any type, optional
        Mask for the dataset. Masks should follow the ``numpy`` convention that
        **valid** data points are marked by ``False`` and **invalid** ones with
        ``True``.
        Defaults to ``None``.

    wcs : any type, optional
        World coordinate system

    meta : `dict`-like object, optional
        Additional meta information about the dataset. If no meta is provided
        an empty `collections.OrderedDict` is created.
        Default is ``None``.

    bbox : BBox, optional
        Bounding Box of the image.

    gain : float, optional
        Gain of the image (e/ADU).  Default is 1.0. 

    rdnoise : float, optional
        Readnoise of the image.  Default is 0.0.

    sky : numpy array, optional
        The sky background array.

    copy : `bool`, optional
        Indicates whether to save the arguments as copy. ``True`` copies
        every attribute before saving it while ``False`` tries to save every
        parameter as reference.
        Note however that it is not always possible to save the input as
        reference.
        Default is ``False``.

    skyfunc : function, optional
        Function that computes the sky background of the image.

    unit : unit-like, optional
        Unit for the dataset. Strings that can be converted to a
        `~astropy.units.Unit` are allowed.
        Default is ``adu``.

    Methods
    -------
    read(fileame)
        ``Classmethod`` to create an CCDData instance based on a ``FITS`` file.
    write(filename)
        Writes the contents of the CCDData instance into a new ``FITS`` file.


    """

    def __init__(self, data, *args, error=None, mask=None, bbox=None, gain=None, rdnoise=None,
                 sky=None,copy=False, skyfunc=None, unit=None, **kwargs):
        # Make sure the original version copies all of the input data
        # otherwise bad things will happen when we convert to native byte-order

        # NDData class has these input parameters
        # data, uncertainty=None, mask=None, wcs=None, meta=None, unit=None, copy=False

        if data.ndim>2:
            raise ValueError('3D CCDData not supported')
        
        # Pull out error from arguments
        if len(args)>0:
            error = args[0]
            if len(args)==1:
                args = ()
            else:
                args = (None,*args[1:])

        # Make sure we have units
        if unit is None:
            unit = 'adu'

        # Check for non-finite values
        bad = (~np.isfinite(data))
        if error is not None:
            bad = bad | (~np.isfinite(error)) | (error <=0)
            if np.sum(bad)>0:
                error[bad] = 1e30
        if np.sum(bad)>0:
            data[bad] = 0.0            
            if mask is None:
                mask = np.array(data.shape,bool)
            mask[bad] = True    # masked means bad
            
        # Initialize with the parent...
        super().__init__(data, *args, mask=mask, copy=copy, unit=unit, **kwargs)

        # Create initial header if none input
        if self.header is None or len(self.header)==0:
            self.header = fits.PrimaryHDU(self.data).header
            # add unit, gain, rdnoise
        
        # Do some checks on error, mask and sky
        if error is not None:
            if error.shape != data.shape:
                raise ValueError('data and error arrays have different shapes')
        if mask is not None:
            if mask.shape != data.shape:
                raise ValueError('data and mask arrays have different shapes')
        if sky is not None:
            if sky.shape != data.shape:
                raise ValueError('data and sky arrays have different shapes')            
        
        # Error
        self._error = error
        # Sky
        self._sky = sky
        # Sky estimation function
        if skyfunc is not None:
            self._skyfunc = skyfunc
        else:
            self._skyfunc = psky.sepsky

        # Gain
        self._gain = gain
        # Read noise
        self._rdnoise = rdnoise
            
        # Copy
        if copy:
            if self._gain is not None:
                self._gain = deepcopy(self._gain)
            if self._rdnoise is not None:
                self._rdnoise = deepcopy(self._rdnoise)
            if self._error is not None:
                self._error = deepcopy(self._error)
            if self._sky is not None:
                self._sky = deepcopy(self._sky)
            self._skyfunc = deepcopy(self._skyfunc)

        # Make Bounding Box
        if bbox is None:
            bbox,x,y = mkbbox(self.data)
            self._bbox = bbox
            self._x = x
            self._y = y
        else:  # bbox input
            self._bbox = bbox
            self._x = None
            self._y = None
            if self.data.ndim==1:
                self._x = np.arange(bbox.xrange[0],bbox.xrange[-1])
            else:
                self._x = np.arange(bbox.xrange[0],bbox.xrange[-1])
                self._y = np.arange(bbox.yrange[0],bbox.yrange[-1])

        
        #self.native()

        # for sep we need to ensure that the data is "c-contiguous"
        # if we used a slice with no copy, then it won't be
        # check image.data.flags['C_CONTIGUOUS']
        # can make it c-contiguous by doing
        # foo = foo.copy(order='C')


        #maybe have CCDData have both a relative and absolute BoundingBox.  The relative being of the
        #last image that it was sliced from (what I'm using now), and the absolute one keeping track of
        #the position in the *original* image.

        
    def __repr__(self):
        prefix = self.__class__.__name__ + '('
        body = np.array2string(self.data, separator=', ', prefix=prefix)
        out = ''.join([prefix, body, ')']) +'\n'
        out += self.bbox.__repr__()
        return out

    def __array__(self):
        """ Return the main data array."""
        return self.data

    # Image arithmetic

    def __add__(self, value):
        newim = self.copy()
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')
            newim.data += value.data
            newim._error = np.sqrt(newim.error**2 + value.error**2)
            # combine masks, get the original by default
            if newim.mask is not None and value.mask is not None:
                newim.mask = np.bitwise_or.reduce((newim.mask,value.mask))
            elif newim.mask is None and value.mask is not None:
                newim.mask = value.mask.copy()
        else:
            newim.data += value
        return newim
        
    def __iadd__(self, value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            self.data += value.data
        else:
            self.data += value
        return self
        
    def __radd__(self, value):
        return self + value
    
    def __sub__(self, value):
        newim = self.copy()
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')
            newim.data -= value.data
            newim._error = np.sqrt(newim.error**2 + value.error**2)
            # combine masks, get the original by default
            if newim.mask is not None and value.mask is not None:
                newim.mask = np.bitwise_or.reduce((newim.mask,value.mask))
            elif newim.mask is None and value.mask is not None:
                newim.mask = value.mask.copy()
        else:
            newim.data -= value
        return newim

    def __isub__(self, value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            self.data -= value.data
        else:
            self.data -= value
        return self
         
    def __rsub__(self, value):
        return self - value 

    def __mul__(self, value):
        newim = self.copy()
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')
            newim.data *= value.data
            newim._error *= value.data
            # combine masks, get the original by default
            if newim.mask is not None and value.mask is not None:
                newim.mask = np.bitwise_or.reduce((newim.mask,value.mask))
            elif newim.mask is None and value.mask is not None:
                newim.mask = value.mask.copy()                        
        else:
            newim.data *= value
            newim._error *= value            
        return newim

    def __imul__(self, value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            self.data *= value.data
            self._error *= value.data
        else:
            self.data *= value
            self._error *= value            
        return self
    
    def __rmul__(self, value):
        return self * value
               
    def __truediv__(self, value):
        newim = self.copy()
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')
            newim.data /= value.data
            newim._error /= value.error
            # combine masks, get the original by default
            if newim.mask is not None and value.mask is not None:
                newim.mask = np.bitwise_or.reduce((newim.mask,value.mask))
            elif newim.mask is None and value.mask is not None:
                newim.mask = value.mask.copy()
        else:
            newim.data /= value
            newim._error /= value            
        return newim

    def __itruediv__(self, value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            self.data /= value.data
            self._error /= value.data
        else:
            self.data /= value
            self._error /= value            
        return self
      
    def __rtruediv__(self, value):
        return self / value

    # Comparison operations
    #  These will all return numpy boolean arrays
    
    def __lt__(self,value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data < value.data
        elif isinstance(value,np.ndarray):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data < value
        else:
            return self.data < value

    def __le__(self,value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data <= value.data
        elif isinstance(value,np.ndarray):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data <= value
        else:
            return self.data <= value    

    def __gt__(self,value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data > value.data
        elif isinstance(value,np.ndarray):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data > value
        else:
            return self.data > value    

    def __ge__(self,value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data <= value.data
        elif isinstance(value,np.ndarray):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data <= value
        else:
            return self.data <= value
    
    def __eq__(self,value):
        if isinstance(value,CCDData):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data == value.data
        elif isinstance(value,np.ndarray):
            if self.shape != value.shape:
                raise ValueError('Shapes do not match')            
            return self.data == value
        else:
            return self.data == value
    
    
    # for the string representation also print out the bbox values
        
    def __getitem__(self, item):
        
        # Abort slicing if the data is a single scalar.
        if self.data.shape == ():
            raise TypeError('scalars cannot be sliced.')

        # BoundingBox input, use its slices
        if isinstance(item,BoundingBox):
            item = item.slices
        # Image input, use its bbox/slices
        if isinstance(item,CCDData):
            item = item.bbox.slices
            
        # Single slice or integer
        #   make sure we have values for each dimension        
        if self.ndim==2 and type(item) is not tuple:
            item = (item,slice(None,None,None))

        # Let the other methods handle slicing.
        kwargs = self._slice(item)        
        new = self.__class__(**kwargs)

        # Deal with error
        if self._error is not None:
            new._error = self._error[item]
        else:
            new._error = None
        # Deal with Sky
        if self._sky is not None:
            new._sky = self._sky[item]
        else:
            new._sky = None
        # Gain and rdnoise
        if self._gain is not None:
            new._gain = deepcopy(self._gain)
        if self._rdnoise is not None:
            new._rdnoise = deepcopy(self._rdnoise)
            
        # Get number of starting values and number of output elements
        # 1-D
        if self.ndim==1:
            nx, = self.data.shape
            # slice object
            if isinstance(item,slice):
                start1,stop1,step1 = item1.indices(nx)                
                nel = stop1-start1
                start = start1
            # Integer
            else:
                nel = 0
                start = item
            newx = self._x[item].copy()
            if nel==0:  # 0-D
                new._x = np.array(newx)
                return new
            new._bbox = BoundingBox(newx[0],newx[-1]+1,0,0)
            new._x = newx
            new._y = None
            
        # 2-D
        elif self.ndim==2:
            shape = self.data.shape
            nel = np.zeros(2,int)
            start = np.zeros(2,int)
            for i in range(2):
                item1 = item[i]
                # Slice object
                if isinstance(item1,slice):
                    start1,stop1,step1 = item1.indices(shape[i])
                    nel[i] = stop1-start1
                    start[i] = start1
                # Integer
                else:
                    nel[i] = 0
                    start[i] = item1
            # python images are (Y,X)
            newx = self._x[item[1]].copy()
            newy = self._y[item[0]].copy()

            # Deal with various output types
            
            # 0-D output
            if np.sum(nel)==0:
                new._x = newx
                new._y = newy
                return new
                
            # 1-D output
            elif np.min(nel)==0:
                rem, = np.where(nel==0)  # which dimension got removed
                if rem[0]==0:
                    new._bbox = BoundingBox(newy[0],newy[-1]+1,0,0)
                    new._x = newy
                    new._y = None
                else:
                    new._bbox = BoundingBox(newx[0],newx[-1]+1,0,0)
                    new._x = newx
                    new._y = None
            # 2-D output
            else:
                try:
                    new._bbox = BoundingBox(newx[0],newx[-1]+1,newy[0],newy[-1]+1)
                except:
                    print('slicing problem')
                    import pdb; pdb.set_trace()
                
                new._x = newx
                new._y = newy

        else:
            raise ValueError('3D CCDData not supported')            
                
        return new

    @property
    def error(self):
        """ Return the uncertainty."""
        # if error not input
        # estimate error from image plus gain
        if hasattr(self,'_error') is False or self._error is None:
            self._error = poissonnoise(self.data,self.gain,self.rdnoise)
        return self._error
    
    @property
    def sky(self):
        """ Return the sky."""
        # estimate the sky
        if self._sky is None:
            self._sky = self._skyfunc(self)
        return self._sky

    @property
    def gain(self):
        """ Return the gain."""
        if self._gain is None:
            self._gain = getgain(self)
        return self._gain

    @property
    def rdnoise(self):
        """ Return the read noise."""
        if self._rdnoise is None:
            self._rdnoise = getrdnoise(self)
        return self._rdnoise    
    
    @property
    def bbox(self):
        """ Boundary box."""
        # Upper values are EXCLUSIVE as is normal in python!
        return self._bbox

    @property
    def x(self):
        """ X-array."""
        return self._x

    @property
    def y(self):
        """ Y-array."""
        return self._y 

    @property
    def min(self):
        """ Calculate the min of the image data.  Uses only unmasked data."""
        if self.mask is not None:
            return np.min(self.data[~self.mask])
        else:
            return np.min(self.data)

    @property
    def max(self):
        """ Calculate the max of the image data.  Uses only unmasked data."""
        if self.mask is not None:
            return np.max(self.data[~self.mask])
        else:        
            return np.max(self.data)

    @property
    def mean(self):
        """ Calculate the mean of the image data.  Uses only unmasked data."""
        if self.mask is not None:
            return np.mean(self.data[~self.mask])
        else:        
            return np.mean(self.data)

    @property
    def median(self):
        """ Calculate the median of the image data.  Uses only unmasked data."""
        if self.mask is not None:
            return np.median(self.data[~self.mask])
        else:        
            return np.median(self.data)

    @property
    def std(self):
        """ Calculate the standard deviation of the image data.  Uses only unmasked data."""
        if self.mask is not None:
            return np.std(self.data[~self.mask])
        else:        
            return np.std(self.data)

    @property
    def mad(self):
        """ Calculate the MAD of the image data.  Uses only unmasked data."""
        if self.mask is not None:
            return dln.mad(self.data[~self.mask])
        else:        
            return dln.mad(self.data)

    def bin(self,binsize,tot=False):
        """ Bin the data in place."""
        if type(binsize) is int:
            binsize = [binsize,binsize]
        nbinpix = binsize[0]*binsize[1]
        # If error does NOT exist yet and not summing, then make it
        # otherwise we cannot use poisson statistics after averaging
        if (hasattr(self,'_error') is False or self._error is None) and tot==False:
            e = self.error   # calculated error behind the scenes
        self.data = dln.rebin(self.data,binsize=binsize,tot=tot)            
        if self._error is not None:
            binerror = dln.rebin(self._error**2,binsize=binsize,tot=True)
            if tot==False:
                binerror /= nbinpix**2
            binerror = np.sqrt(binerror)
            self._error = binerror
        if self.mask is not None:
            newmask = dln.rebin(self.mask.astype(int),binsize=binsize,tot=True)
            newmask = (newmask > 0.5*nbinpix)
            self.mask = newmask
        if self._sky is not None:
            self._sky = dln.rebin(self._sky,binsize=binsize,tot=tot)
        # Update header
        self.header['NAXIS1'] = self.data.shape[1]
        self.header['NAXIS2'] = self.data.shape[0]
        self.header['XBIN'] = binsize[0]
        self.header['YBIN'] = binsize[1]
        if 'CRPIX1' in self.header:
            self.header['CRPIX1'] /= binsize[0]
        if 'CRPIX2' in self.header:
            self.header['CRPIX2'] /= binsize[1]
        if 'CDELT1' in self.header:
            self.header['CDELT1'] *= binsize[0]
        if 'CDELT2' in self.header:
            self.header['CDELT2'] *= binsize[1]
        if 'CD1_1' in self.header:
            self.header['CD1_1'] *= binsize[0]
        if 'CD1_2' in self.header:
            self.header['CD1_2'] *= binsize[1]
        if 'CD2_1' in self.header:
            self.header['CD2_1'] *= binsize[0]
        if 'CD2_2' in self.header:
            self.header['CD2_2'] *= binsize[1]
            
        # Update BoundingBox
        bbox,x,y = mkbbox(self.data)
        self._bbox = bbox
        self._x = x
        self._y = y
        # Fix wcs
        #  crval  keep
        #  crpix  scale by binsize
        #  cd/cdelt  scale by binsize
        #  distortion terms,  scale by binsize with the appropriate power
        if self.wcs is not None:
            if self.wcs.wcs.ctype[0] != '':
                self.wcs.wcs.crpix /= np.array(binsize)
                if hasattr(self.wcs.wcs,'cd'):
                    self.wcs.wcs.cd[:,0] *= binsize[0]
                    self.wcs.wcs.cd[:,1] *= binsize[1]     
                else:
                    #self.wcs.wcs.pc[:,0] *= binsize[0]
                    #self.wcs.wcs.pc[:,1] *= binsize[1]                     
                    self.wcs.wcs.cdelt *= np.array(binsize)
            # something still isn't quite right
            self.wcs.array_shape = self.shape
        
    def resetbbox(self):
        """ Forgot the original coordinates in BoundingBox."""
        self.bbox.reset()
        
    def isnative(self,data):
        """ Check if data has native byte order."""
        sys_is_le = sys.byteorder == 'little'
        native_code = sys_is_le and '<' or '>'
        # = is native
        # | is for not applicable
        return (data.dtype.byteorder == native_code) or (data.dtype.byteorder=='=') or (data.dtype.byteorder=='|')

    def isccont(self,data):
        """ Check if data is c-continuous."""
        return data.flags['C_CONTIGUOUS']

    def issepready(self,data):
        """ Check if data is ready for sep (native byte order and c-continuous)."""
        return self.isnative(data) & self.isccont(data)
    
    def native(self):
        """ Make sure that the arrays use native endian for sep."""

        # Deal with byte order for sep
        sys_is_le = sys.byteorder == 'little'
        native_code = sys_is_le and '<' or '>'
        # data
        if self.data.dtype.byteorder != native_code:
            self.data = self.data.byteswap(inplace=True).newbyteorder()
        # error
        if self._error is not None:
            if self._error.dtype.byteorder != native_code:
                self._error = self._error.byteswap(inplace=True).newbyteorder()
        # mask
        if self.mask is not None:
            if self.mask.dtype.byteorder != native_code:
                self.mask = self.mask.byteswap(inplace=True).newbyteorder()
        # sky
        if self._sky is not None:
            if self._sky.dtype.byteorder != native_code:
                self._sky = self._sky.byteswap(inplace=True).newbyteorder()            

    @property
    def ccont(self):
        """ Return C-Continuous data for data, error, mask, sky."""

        if self.mask is not None:
            return (self.sepready(self.data), self.sepready(self.error),
                    self.sepready(self.mask), self.sepready(self.sky))
        else:
            return (self.sepready(self.data), self.sepready(self.error),
                    None, self.sepready(self.sky))
            
        #if self.data.flags['C_CONTIGUOUS']==False:
        #    data = self.data.copy(order='C')
        #else:
        #    data = self.data
        #if self.error.flags['C_CONTIGUOUS']==False:
        #    error = self.error.copy(order='C')
        #else:
        #    error = self.error
        #if self.mask is not None:
        #    if self.mask.flags['C_CONTIGUOUS']==False:
        #        mask = self.mask.copy(order='C')
        #    else:
        #        mask = self.mask
        #else:
        #    mask = self.mask
        #if self.sky.flags['C_CONTIGUOUS']==False:
        #    sky = self.sky.copy(order='C')
        #else:
        #    sky = self.sky

        return (data,error,mask,sky)

    def sepready(self,data=None):
        """ Return sep-ready data (native byte order and c-continuous)."""

        # No data nput, return a sep-ready version of the
        #   the image object
        if data is None:
            # Check all data arrays
            ready = True
            for n in ['data','_error','mask','_sky']:
                dat = getattr(self,n)
                if dat is not None:
                    ready &= self.sepready(dat)
            # Already sep-ready
            if ready:
                return self
            # Not ready, get sep-ready versions of the data
            new = self.copy()
            for n in ['data','_error','mask','_sky']:
                dat = getattr(new,n)
                if dat is not None:
                    setattr(new,n,new.sepready(dat))
            return new
        
        # Data array input
        else:
            if self.issepready(data)==False:
                new = data.copy(order='C')
                if self.isnative(new)==False:
                    new = new.byteswap(inplace=True).newbyteorder()
                return new
                #return data.copy(order='C').byteswap(inplace=True).newbyteorder()
            else:
                return data
                
    @property
    def sepdata(self):
        """ Return C-Continuous and native byte order data for sep."""

        # Deal with byte order for sep
        sys_is_le = sys.byteorder == 'little'
        native_code = sys_is_le and '<' or '>'
        
        # Loop over data types
        out = []
        for name in ['data','error','mask','_sky']:
            data = getattr(self,name)
            if data is not None:
                # if not c-contiguous or not native byte order, copy + correct
                if data.flags['C_CONTIGUOUS']==False or data.dtype.byteorder!=native_code:
                    out.append(data.copy(order='C').byteswap(inplace=True).newbyteorder())
                else:
                    out.append(data)
            else:
                out.append(data)
        return tuple(out)
            
    def copy(self):
        """
        Return a copy of the CCDData object.
        """
        return self.__class__(self, copy=True, error=self._error, gain=self._gain, rdnoise=self._rdnoise)

    def write(self,outfile,overwrite=True):
        """
        Write the image data to a file.

        Parameters
        ----------
        outfile : str
          The output filename.
        overwrite : boolean, optional
          Overwrite the file if it already exists.  Default is True.

        Example
        -------

        image.write('image.fits',overwrite=True)

        """

        hdulist = fits.HDUList()
        # HDU0: Data and header
        hdulist.append(fits.PrimaryHDU(self.data,self.header))
        hdulist[0].header['IMAGTYPE'] = 'Prometheus'
        hdulist[0].header['BUNIT'] = 'Flux'
        # HDU1: error
        hdulist.append(fits.ImageHDU(self.error))
        hdulist[1].header['BUNIT'] = 'Uncertainty'
        # HDU2: mask
        if self.mask is None:
            hdulist.append(fits.ImageHDU(self.mask))
        else:
            hdulist.append(fits.ImageHDU(self.mask.astype(int)))            
        hdulist[2].header['BUNIT'] = 'Mask'
        # HDU3: flags
        hdulist.append(fits.ImageHDU(self.flags))
        hdulist[3].header['BUNIT'] = 'Flags'
        # HDU4: sky
        hdulist.append(fits.ImageHDU(self.sky))
        hdulist[4].header['BUNIT'] = 'Sky'
        hdulist.writeto(outfile,overwrite=overwrite)
        hdulist.close()

    def tohdu(self):
        """
        Convert the image to an HDU so it can be written to a file.
        Note that only the image data is converted to the HDU
        (no error, mask, flags or sky).

        Returns
        -------
        hdu : fits HDU object
          The FITS HDU object.

        Example
        -------

        hdu = image.tohdu()

        """
        
        # Data and header
        if len(self.header)>0:
            hdu = fits.PrimaryHDU(self.data,self.header)
        else:
            hdu = fits.PrimaryHDU(self.data)
        hdu.header['IMAGTYPE'] = 'Prometheus'        
        return hdu
        
    @classmethod
    def read(cls,filename):
        """ Read in an image from file."""

        hdulist = fits.open(filename)
        nhdu = len(hdulist)
        # Checking if there's an image in HDU0 or HDU1
        if nhdu==1 and hdulist[0].data is None:
            print('No image in '+filename)
        if nhdu>1 and hdulist[0].data is None and hdulist[1].data is None:
            print('No image in first two extensions of '+filename)            
        # Prometheus file type
        if hdulist[0].header.get('IMAGTYPE') == 'Prometheus':
            # HDU0: Data and header
            data = hdulist[0].data
            head = hdulist[0].header
            # HDU1: error
            if nhdu>1:
                error = hdulist[1].data
                ehead = hdulist[1].header
            else:
                error = None
            # HDU2: mask
            if nhdu>2:
                mask = hdulist[2].data
                mhead = hdulist[2].header
            else:
                mask = None
            # HDU3: flags
            if nhdu>3:
                flags = hdulist[3].data
                fhead = hdulist[3].header
            else:
                flags = None
            # HDU4: sky
            if nhdu>4:
                sky = hdulist[4].data
                shead = hdulist[4].header
            else:
                sky = None
        # Other file type
        else:
            datanum = 0
            if hdulist[0].data is not None: datanum=1
                
            # Image in HDU0
            if hdulist[0].data is not None:
                # data
                data = hdulist[0].data
                head = hdulist[0].header
                # error
                if nhdu>1 and hdulist[1].data is not None:
                    error = hdulist[1].data
                else:
                    error = None
            else:
                # data
                data = hdulist[1].data
                head = hdulist[1].header
                # error
                if nhdu>2 and hdulist[2].data is not None:
                    error = hdulist[2].data
                else:
                    error = None
            # mask
            mask = np.zeros(data.shape,bool)
            bad = (~np.isfinite(data))
            if error is not None:
                bad = bad | (~np.isfinite(error))
            mask[bad] = True    # masked means bad
            # flags, sky
            flags = None
            sky = None
            
        hdulist.close()
        # Make WCS, this doesn't capture the PV#_# values
        w = WCS(head)
        # Units
        unit = head.get('bunit')
        if unit=='ADU':
            unit = 'adu'
        if unit is None:
            unit = 'adu'

        # Make sure the data is float
        data = data.astype(float)
            
        # make the ccddata object
        image = CCDData(data,error=error,mask=mask,meta=head,
                        flags=flags,sky=sky,wcs=w,unit=unit)
        
        return image

    
class BoundingBox(BBox):

    def __init__(self, *args, **kwargs):

        # Initialize with the parent...
        super().__init__(*args, **kwargs)

    @property
    def xrange(self):
        return (self.ixmin,self.ixmax)

    @property
    def yrange(self):
        return (self.iymin,self.iymax)    
        
    @property
    def data(self):
        return [(self.ixmin,self.ixmax),(self.iymin,self.iymax)]
        
    def __getitem__(self,item):
        return self.data[item]

    def __array__(self):
        return np.array(self.data)

    @property
    def slices(self):
        """ Return the slices."""
        return (slice(self.iymin,self.iymax,None),
                slice(self.ixmin,self.ixmax,None))

    def reset(self):
        """ Forget the original coordinates."""
        self.ixmax -= self.ixmin
        self.iymax -= self.iymin
        self.ixmin = 0
        self.iymin = 0
